/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  ngl.h
\brief Common declarations

This file must be included \e first in every source file, in NGL own sources
and in NGL user application code.
*/

#ifndef __ngl_h__
#define __ngl_h__


/*
 * Import NGL build settings
 */

#include "ngl_config.h"


/*
 * Platform specific includes
 */

/* Win32
 */
#ifdef _WIN32_
//#define _WINSOCKAPI_  // Prevent inclusion of winsock.h in windows.h
  #include <windows.h>
  #include <malloc.h>
  #include <cstdio>

  #pragma warning(disable : 4786) // Disable the "debug name too long" warning which is really annoying with stlport
  #pragma warning(disable : 4251) // Disable the template export warning
  #pragma warning(disable : 4530) // Disable the exception disable but used in MSCV Stl warning.

//#include <winsock2.h>
  #ifdef _NO_NOT_USED_PARAM_WARNING_
  #pragma warning (disable: 4100) // Prevent the "unreferenced formal parameter" warning
  #pragma warning (disable: 4201) // Prevent the "parameter not used" warning
  #pragma warning (disable: 4711) // Prevent the "selected for automatic inline expansion" warning
  #endif
  
  #ifdef _MAYAPLUGIN_ // Maya includes try to redefine the bool type
    #define _BOOL
  #endif

  #define WM_NGLTIMER (WM_USER+10)

  // Fixed size types
  typedef signed char    int8;
  typedef unsigned char  uint8;
  typedef short          int16;
  typedef unsigned short uint16;
  typedef int            int32;
  typedef unsigned int   uint32;
  typedef __int64        int64;
  typedef unsigned __int64 uint64;

  // Windows already use these macros for Unicode support
  // (we keep their expected behaviour)
  #undef _T
  #undef TEXT
#endif // _WIN32_


/* Unices
 */
#ifdef _UNIX_
  #ifndef _GNU_SOURCE
  #define _GNU_SOURCE
  #endif
  #include <stdlib.h>
  #include <sys/types.h>
  #include <ctype.h>

  typedef int8_t    int8;
  typedef int16_t   int16;
  typedef int32_t   int32;
  typedef int64_t   int64;
  typedef u_int8_t  uint8;
  typedef u_int16_t uint16;
  typedef u_int32_t uint32;
  typedef u_int64_t uint64;
#endif // _UNIX_


/* MacOS/Carbon
 */
#ifdef _CARBON_
  #ifdef __MWERKS__
    // Codewarrior has another view of how the headers should be organised
    #if macintosh == 0
      #include <Carbon/Carbon.h>
      #include <stdlib.h>
      #include <sys/types.h>

      // This is ugly but the is a name conflict in codewarrior in between BSD and MSL headers
      #define int32_t ::int32_t
    #else
      #include <Carbon.h>
    #endif
    #include "alloca.h"
  #else
    #include <carbon/carbon.h>
  #endif // __MWERKS__
  #include <stdlib.h>

  typedef int8_t    int8;
  typedef int16_t   int16;
  typedef int32_t   int32;
  typedef int64_t   int64;
  typedef u_int8_t  uint8;
  typedef u_int16_t uint16;
  typedef u_int32_t uint32;
  #if macintosh == 1
    typedef UInt64 uint64;
  #else
    typedef u_int64_t uint64;
  #endif

  #ifdef __MWERKS__
    #define PATH_MAX 1024
  #endif
#endif // _CARBON_

  
/*
 * Common stuff for all platforms
 */

//! Symbol linkage info, necessary for some compilers like MSVC (empty for the others)
#ifndef NGL_API
#define NGL_API
#endif

//! Shortcut and uniformisation with fixed-size NGL integers
typedef unsigned int uint;

//! File offset and size type
/*!
This type is used for file adressing (size and offsets). It is currently
set to a 32 bit signed integer (can address 2GB files on filesystems which
allow it).
*/
#ifdef _UNIX_
  #include <unistd.h>
  typedef ssize_t nglSize;
#else
  typedef int nglSize;
#endif // _UNIX_

#ifdef _T
  // _T is defined under MacOSX
  #undef _T
#endif

#if defined(USE_WCHAR) || defined(_DOXYGEN_)
  typedef wchar_t nglChar;
  /*!< Portable char type
    See nglString for more info
  */
  #define _T(x) L##x
  /*!< Portable string declaration macro
    Use for literals to support both <i>char</i> and <i>wide char</i> modes :
    \code
    const nglChar* my_string_value = _T("Example");
    nglString my_string = my_string_value;
    nglString my_string(_T("Example"));
    \endcode
  */
#else
  typedef char nglChar;
  #define _T(x) x
#endif

#define TEXT(x) _T(x)
/*!< Portable string declaration macro
  An alternative to the _T macro
*/


/* Handy macros
 */
#if (!defined _BEOS_)
#  define MIN(a,b) ((a) < (b) ? (a) : (b))
#  define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif

#ifdef UNUSED
#  undef UNUSED
#endif
#define UNUSED(x)    ((x) = (x))

/* At last, include GL and GLU headers if necessary
 */
#ifndef _NOGFX_
#  ifdef _CARBON_
#    if defined(__MWERKS__) && macintosh == 0
#      include <AGL/agl.h>
#      include <OpenGL/glu.h>
#    else
#      include <agl.h>
#      include <glu.h>
#    endif
#  else
#    include <GL/gl.h>
#    include <GL/glu.h>
#  endif
#  ifdef _WIN32_
     /* The latest glext.h file can be fetched from :
      *   http://oss.sgi.com/projects/ogl-sample/registry/index.html
      */
#    include <GL/glext.h>
#  endif
#endif // _NOGFX_

// This linkage macro should be defined by <GL/gl.h> if needed
#ifndef APIENTRY
#define APIENTRY
#endif


/*
 * NGL macros
 */

/* Log domains
 *
 * Keep in sync with the declaration in src/core/nglKernel_shr.cpp
 */
#define NGL_DOMAIN_DEFAULT     ((uint) 0)
#define NGL_DOMAIN_KERNEL      ((uint) 1)
#define NGL_DOMAIN_APP         ((uint) 2)
#define NGL_DOMAIN_STRINGCONV  ((uint) 3)
#define NGL_DOMAIN_CON         ((uint) 4)
#define NGL_DOMAIN_TIMER       ((uint) 5)
#define NGL_DOMAIN_MATH        ((uint) 6)
#define NGL_DOMAIN_PATH        ((uint) 7)
#define NGL_DOMAIN_FILE        ((uint) 8)
#define NGL_DOMAIN_STREAM      ((uint) 9)
#define NGL_DOMAIN_FONT        ((uint)10)
#define NGL_DOMAIN_NET         ((uint)11)
#define NGL_DOMAIN_IMAGE       ((uint)12)
#define NGL_DOMAIN_MODULE      ((uint)13)
#define NGL_DOMAIN_INPUT       ((uint)14)
#define NGL_DOMAIN_3DS         ((uint)15)
#ifdef _NOGFX_
#  define NGL_DOMAIN_LAST      ((uint)15)
#else
#  define NGL_DOMAIN_WIN       ((uint)16)
#  define NGL_DOMAIN_CONTEXT   ((uint)17)
#  define NGL_DOMAIN_VMODE     ((uint)18)
#  define NGL_DOMAIN_LAST      ((uint)18)
#endif //_NOGFX_

/* Verbosity levels
 */
#define NGL_LOG_ALWAYS  (0)
#define NGL_LOG_ERROR   (1)
#define NGL_LOG_WARNING (2)
#define NGL_LOG_INFO    (3)
#define NGL_LOG_DEBUG   (8)

/* Default verbose level (for all domains)
 */
#ifndef LOG_LEVEL
#  ifdef _DEBUG_
#    define LOG_LEVEL NGL_LOG_INFO
#  else
#    define LOG_LEVEL NGL_LOG_ERROR
#  endif
#endif

#endif // __ngl_h__
