/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglContext.h
\brief OpenGL context management

These classes is not available if the _NOGFX_ symbol is defined.
*/

#ifndef __nglContext_h__
#define __nglContext_h__

#include "ngl.h"

#ifndef _NOGFX_

#include "nglError.h"
#include "nglKernel.h"

#ifdef _X11_
#include <X11/Xutil.h>
#ifndef GLX_GLXEXT_LEGACY // !+rzr @ Linux/Debian + Custom XServer (inc Mesa5)
#define GLX_GLXEXT_LEGACY // !+rzr @ needed coz' function glx...ARB.. was Undef
#endif // GLX_GLXEXT_LEGACY // !+rzr @ Not sure if it is good or not, !!!CHECK
#include <GL/glx.h>
#endif // _X11_


//! OpenGL context description
/*!
This simple container stores the different properties of a given GL context
and provide some helpers to fill in the values.
*/
class NGL_API nglContextInfo
{
public:
  uint FrameCnt;     ///< Number of frame buffers (two means double-buffering)
  uint FrameBitsR;   ///< Bits per red component (frame buffer)
  uint FrameBitsG;   ///< Bits per green component (frame buffer)
  uint FrameBitsB;   ///< Bits per blue component (frame buffer)
  uint FrameBitsA;   ///< Bits per alpha component (frame buffer)
  uint DepthBits;    ///< Depth buffer resolution (ie. Z buffer, 0 means no Z buffer)
  uint StencilBits;  ///< Stencil buffer resolution (0 means no stencil)
  uint AccumBitsR;   ///< Bits per red component (accumulator buffer)
  uint AccumBitsG;   ///< Bits per green component (accumulator buffer)
  uint AccumBitsB;   ///< Bits per blue component (accumulator buffer)
  uint AccumBitsA;   ///< Bits per alpha component (accumulator buffer)
  uint AuxCnt;       ///< Number of auxiliary buffers
  uint AABufferCnt;  ///< Number of anti-aliasing buffers
  uint AASampleCnt;  ///< Anti-alisaing oversampling count
  bool Stereo;       ///< Stereoscopic display

  /** @name Life cycle */
  //@{
  nglContextInfo();
  /*!<
    Build a 'reasonable' default GL context description.

    The context info shows the following properties :

    - RGBA color model (not indexed mode)
    - the frame buffer has the current desktop's bitdepth 
    - double buffering or more
    - at least 16 bits of depth buffer
    - main plane layer (not overlay, not underlay)
  */
  nglContextInfo(const nglContextInfo& rInfo);  ///< Copy constructor
  //@}

  /** @name Logging */
  //@{
  void Dump(int Level) const;  ///< Dumps human readable context information using verbosity level \a Level
  //@}

  /** @name Context enumeration */
  //@{
  static bool Enum (uint Index, nglContextInfo& rInfo);
  /*!< Enumerate supported GL contexts
    \param Index zero-based index
    \param rInfo context description is written here
    \return if true, \a rInfo contains valid informations

    Example :

\code
void DumpContexts()
{
  nglContextInfo info;
  uint index = 0;

  NGL_LOG(NGL_DOMAIN_CONTEXT, NGL_LOG_INFO, _T("Enumerating available GL contexts :"));
  while (nglContextInfo::Enum(index++, info))
    info.Dump(NGL_LOG_INFO);
}
\endcode

    \b Note : on multiple display systems, the enumeration is made on the
    default/main display device.
  */
  //@}

private:
#ifdef _WIN32_
  int mPFD;

  nglContextInfo (HDC hDC, int PFD);

  int GetPFD(HDC hDC) const;
#endif // _WIN32_

#ifdef _X11_
  static XVisualInfo* mpXVisualInfoList;
  static uint         mXVisualInfoListCnt;

  static bool         Init();
  static void         Exit();

  XVisualInfo* mpXVisualInfo;

  nglContextInfo (Display* pDisplay, XVisualInfo* pXVisualInfo);

  XVisualInfo* GetXVisualInfo (Display* pDisplay, int Screen) const;
#endif // _X11_

  friend class nglContext;
};


#define NGL_CONTEXT_ENONE  0 ///< No error


//! OpenGL context management
/*!
This is an abstract class which stores per GL context informations for a
rendering surface such as nglWindow. This is also the place from where GL
extensions are reachable. You can query and initialize an extension with
checkExtension() and use extensions functions as if they were methods of
this context (they are really pointers to functions which are initially
NULL).

Developer's note : the extension support code is generated by a Perl script
which parses <em>glext.h</em>, see <em>src/core/glext</em> in the source
distribution for more info.
*/
class NGL_API nglContext : public nglError
{
public:
  /** @name GL extensions */
  //@{
  bool CheckExtension(const nglChar* pExtName);
  /*!< Check presence of a GL extension and initialize it if found
    \param pExtName GL extension name
    \return true if extension present and ready

    This method checks if a GL extension is supported (by both client and
    server sides) and prepares it for use. Example :

\code
if (HasExtension(_T("GL_ARB_texture_compression")))
{
  GLint value;
  glGetIntegerv(GL_NUM_COMPRESSED_TEXTURE_FORMATS_ARB, &value);
  NGL_OUT(_T("%d texture compression formats supported), value);
}
\endcode

    Note that you need an up-to-date <GL/glext.h> file from your OpenGL
    implementation.
    For the full extension list and documentation, see the centralized
    registry at http://oss.sgi.com/projects/ogl-sample/registry/
  */
  //@}

  // Include extension methods as members here
  #include "ngl_glext.h"

  /** @name  */
  //@{
  bool GetContextInfo(nglContextInfo& rInfo) const;  ///< Retrieve context description
  void Dump(int Level) const;
  //@}

protected:
  /** @name Life cycle */
  //@{
  nglContext();  ///< Build an unitialized context
  virtual ~nglContext();
  //@}

  virtual const nglChar* OnError (uint& rError) const;

private:
  typedef void (*GLExtFunc)(void);

  bool      InitExtension (const char* pExtName);
  GLExtFunc LookupExtFunc (const char* pFuncName);

#ifdef _WIN32_
protected:
  HDC    mDC;
  HGLRC  mRC;

  bool Build(HWND hWnd, const nglContextInfo& rInfo, const nglContext* pShared);
  bool MakeCurrent() const;

private:
  HWND   mCtxWnd;
#endif // _WIN32_

#ifdef _X11_
protected:
  Display*     mpDisplay;
  XVisualInfo* mpXVisualInfo;
  Visual*      mpXVisual;
  GLXContext   mGlxCtx;
  int          mDepth;
  int          mGlxErrorBase, mGlxEventBase;

  bool Build(int Screen, const nglContextInfo& rInfo, const nglContext* pShared);
  bool MakeCurrent(Window Win) const;
#endif // _X11_

#ifdef _CARBON_
  protected:
  AGLPixelFormat mFormat;
  AGLContext     mCtx;
  bool Build(WindowRef Win, const nglContextInfo& rInfo, const nglContext* pShared);
  bool MakeCurrent(WindowRef Win) const;
    
  CFBundleRef mpBundleRefOpenGL;
  OSStatus aglInitEntryPoints (void);
  void aglDellocEntryPoints (void);
  void* aglGetProcAddress (const char * pszProc);
#endif
};

#endif // !_NOGFX_
#endif // __nglContext_h__
