/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglFile.h
\brief File input/output abstraction
*/

#ifndef __nglFile_h__
#define __nglFile_h__

#include "ngl.h"
#include "nglError.h"
#include "nglEndian.h"
#include "nglPath.h"
#include "nglString.h"


// Error codes
#define NGL_FILE_ENONE     0
#define NGL_FILE_EISDIR    1
#define NGL_FILE_EACCESS   2
#define NGL_FILE_ETOOLONG  3
#define NGL_FILE_EBADPATH  4
#define NGL_FILE_ERONLY    5
#define NGL_FILE_ENOSPACE  6
#define NGL_FILE_EPROCMAX  7
#define NGL_FILE_ESYSMAX   8


//! File opening mode
/*!
The file opening mode is set from the nglFile constructor invocation.
*/
enum nglFileMode
{
  eFileRead,   /*!< The file can be read but not written to.
                    The file must exist. */
  eFileWrite,  /*!< The file can be written to (and read).
                    If the file already exists, it is erased (length truncated to 0), otherwise it is created. */
  eFileModify, /*!< The file can be written to (and read).
                    If the file already exists, its content is preserved, otherwise it is created.
                    The read/write position is initially set to the beginning of the file. */
  eFileAppend  /*!< The file can be written to (and read).
                    If the file already exists, its content is preserved, otherwise it is created.
                    The read/write position is initially set to the end of the file. */
};

//! Seek hint
/*!
Seeking positions and directions for nglFile::SetPos().
*/
enum nglFileWhence
{
  eFileFromStart,  ///< Move forward, from the beginning of the file (offset 0).
  eFileForward,    ///< Move forward from the current read/write position.
  eFileRewind,     ///< Move backwards from the current read/write position.
  eFileFromEnd     ///< Move backwards from the end of the file.
};


//! Portable file class
/*!
To open a file, you only need to instantiate a nglFile object. You will need a
valid nglPath object first. You can either use the nglPath::Open method (this
can be desirable if you need file polymorphism) or the nglFile constructor.

You can set explicitely the file endianness from the constructor. If the application
is built on a platform with a different endianness from the one you specify, proper
byte swapping will occur automatically on word access (read/write).

To close a file (needed to properly finalize write access), you muste delete the
file object.

File are buffered by the system as a default : the Write* methods return immediatly
but the data is actually written later on the medium. You can force all data writes
pending on a currently opened file with the Flush() method (this will have no effect
on read-only files). You can also unbuffer the file with the AutoFlush() method :
setting auto-flush to true will make all write access immediate (ie. Write* methods
will only return when the data has been written on the medium).
*/
class NGL_API nglFile : public nglError
{
public:
  /** @name Life cycle */
  //@{
  nglFile (const nglPath& rPath, nglFileMode Mode = eFileRead, nglEndian Endian = eEndianNative);
  /*!<
    \param rPath file's path
    \param Mode read/write mode
    \param Endian endianness convention

    Open the file designated by \a rPath, using read/write \a Mode. The file format can be
    set to a specific endianness, build platform's one being used as a default.

    You must check any errors with GetError() after calling the constructor.
    If there is an error, the object is unusable and should be destroyed.
  */
  virtual ~nglFile();
  /*!< Close the file (and flush pending writes if any). */
  //@}

  /** @name Properties */
  //@{
  nglPath GetPath() const;
  /*!< Retrieve file's path
    \return a copy of the path passed to the constructor
  */
  nglFileMode GetMode() const;
  /*!< Get file's read/write mode
    \return the file mode passed to the constructor
  */
  nglEndian GetEndian() const;
  /*!< Get file's endianness convention
    \return current endianness convention
  */
  void SetEndian(nglEndian Endian);
  /*!<
    \param Endian new endianness convention

    Set file's endianness on the fly. All further read and writes will use the
    new endianness (buffered pending writes are not affected).
  */
  nglSize GetSize() const;
  /*!< Get current file size
    \return current (real time, including last writes) file size in bytes
  */
  //@}

  /** @name Data flow */
  //@{
  bool IsOpen() const;
  /*!< Check if the file is open
    \return true if the file is open

    Some methods can set an an error for a nglFile object (see GetError()), however
    the file is still open (and thus usable). This method let you tell if an error
    was fatal or not.

    Note : errors set by the nglFile constructors are \e all fatal (ie. IsOpen()
    will return \e false if an error was set).
  */
  bool IsEOF() const;
  /*!< Check End Of File status
    \return \e true if the read/write position reached the end of the file (no more bytes to read).
  */
  nglSize GetPos() const;
  /*!< Return current read/write position
    \return current read/write position in byte units
  */
  nglSize SetPos(nglSize Offset, nglFileWhence Whence = eFileFromStart);
  /*!<
    \param Offset relative offset in byte units
    \param Whence displacement origin and direction

    Set the current read/write position, in byte units. \a Offset must be positive,
    the direction can be choosen using the proper nglFileWhence item.
  */
  void Flush();
  /*!<
    Immediatly put pending writes on the storage medium (ignored for a read-only file)
  */
  void AutoFlush(bool Auto = true);
  /*!<
    \param Auto turn on/off auto flushing

    If auto-flush is set, all write operations will return only once the data is
    effectively written on the storage medium
  */
  //@}

  /** @name Data reading */
  //@{
  nglSize ReadInt8   (uint8*  pData, nglSize Count = 1);  ///< Read \a Count 8-bit integer values, returns number of bytes effectively read
  nglSize ReadInt16  (uint16* pData, nglSize Count = 1);  ///< Read \a Count 16-bit integer values, returns number of 16-bit words effectively read
  nglSize ReadInt32  (uint32* pData, nglSize Count = 1);  ///< Read \a Count 32-bit integer values, returns number of 32-bit words effectively read
  nglSize ReadInt64  (uint64* pData, nglSize Count = 1);  ///< Read \a Count 64-bit integer values, returns number of 64-bit words effectively read
  nglSize ReadFloat  (float*  pData, nglSize Count = 1);  ///< Read \a Count float values (IEEE 754 simple precision, 32 bits), returns number of 32-bit words effectively read
  nglSize ReadDouble (double* pData, nglSize Count = 1);  ///< Read \a Count double values (IEEE 754 double precision, 64 bits), returns number of 64-bit words effectively read

  virtual nglSize Read (void* pData, nglSize WordCount, uint WordSize);  ///< Read \a WordCount number of \a WordSize bytes long into \a pBuffer, returns the number of words effectively read

  uint8   ReadInt8   (uint8 Default = 0);     ///< Read an 8-bit integer value, returns \a Default if no data is available
  uint16  ReadInt16  (uint16 Default = 0);    ///< Read a 16-bit integer value, returns \a Default if no data is available
  uint32  ReadInt32  (uint32 Default = 0);    ///< Read a 32-bit integer value, returns \a Default if no data is available
  uint64  ReadInt64  (uint64 Default = 0);    ///< Read a 64-bit integer value, returns \a Default if no data is available
  float   ReadFloat  (float Default = 0.0);   ///< Read a float value (IEEE 754 simple precision, 32 bits), returns \a Default if no data is available
  double  ReadDouble (double Default = 0.0);  ///< Read a double value (IEEE 754 double precision, 64 bits), returns \a Default if no data is available
  //@}

  /** @name Data writing */
  //@{
  nglSize WriteInt8   (const uint8*  pData, nglSize Count = 1);  ///< Write \a Count 8-bit integer values, returns number of bytes effectively written
  nglSize WriteInt16  (const uint16* pData, nglSize Count = 1);  ///< Write \a Count 16-bit integer values, returns number of 16-bit words effectively written
  nglSize WriteInt32  (const uint32* pData, nglSize Count = 1);  ///< Write \a Count 32-bit integer values, returns number of 32-bit words effectively written
  nglSize WriteInt64  (const uint64* pData, nglSize Count = 1);  ///< Write \a Count 64-bit integer values, returns number of 64-bit words effectively written
  nglSize WriteFloat  (const float*  pData, nglSize Count = 1);  ///< Write \a Count float values (IEEE 754 simple precision, 32 bits), returns number of 32-bit words effectively written
  nglSize WriteDouble (const double* pData, nglSize Count = 1);  ///< Write \a Count double values (IEEE 754 double precision, 64 bits), returns number of 64-bit words effectively written

  virtual nglSize Write (const void* pData, nglSize WordCount, uint WordSize);  ///< Write \a WordCount number of \a WordSize bytes long from \a pBuffer, returns the number of words effectively written
  //@}

protected:
  virtual const nglChar* OnError (uint& rError) const;

private:
  nglPath         mPath;
  nglFileMode     mMode;
  nglEndian       mEndian;
  bool            mAutoFlush;
#if defined(_UNIX_) || defined(_CARBON_)
  int             mFD;
#endif

#ifdef _WIN32_
  HANDLE mFD;
#endif
};

#endif // __nglFile_h__
