/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglFont1.h
\brief Font manager and renderer (DEPRECATED)
*/

#ifndef __nglFont1_h__
#define __nglFont1_h__

#include "ngl.h"
#include "nglApplication.h"
#include "nglError.h"
#include "nglPath.h"
#ifdef USE_FREETYPE
#  include <ft2build.h>
#  include FT_FREETYPE_H
#  include FT_CACHE_IMAGE_H
#endif // USE_FREETYPE
#include <map>


#define NGL_FONT1_DEFAULT_SIZE     (12.0f)
#define NGL_FONT1_PRESERVE_DEFAULT (GL_COLOR_BUFFER_BIT | GL_CURRENT_BIT | GL_ENABLE_BIT | GL_TEXTURE_BIT)


//! Character maps
/*!
A character map let you convert a character value from a given encoding (Unicode,
ISO8859-*, ASCII, etc) to the corresponding font glyph. A character map is named
after the encoding it expects as input. Hence it sets the encoding expected
by the nglFont1::Print method.

To make proper use of text encoding and font character maps, you should :

- Select your desired character map. In most cases, the default map of a nglFont1
  is ASCII compatible (such as Unicode or ISO8859-*), meaning you can access the
  standard ASCII printable characters from 33 (!) through 126 (~).

- Check your requested character map is available (see nglFont1::HasCharMap) or
  has been applied (see nglFont1::GetCharMap and nglFont1::SetCharMap)

Note: NGL does not have proper i18n support for strings, so the current implementation
only truly works for 8bit encodings. Unicode support is on the way.

More info on encodings :
- http://czyborra.com/charsets/
- http://www.unicode.org/Public/MAPPINGS/
*/
enum nglCharMap
{
  eCharMapNone,            ///< No charmap. This situation occurs if font loading failed.

  eCharMapSymbol,          ///< Adobe Symbol
  eCharMapUnicode,         ///< Unicode (UTF-32) mapping
  eCharMapLatin2,          ///< Latin 2 (ISO-8859-2, East European languages)
  eCharMapSJIS,            ///< ShiftJIS (Japanese)
  eCharMapGB2312,          ///< GB-2312 (Chinese)
  eCharMapBig5,            ///< Big5 (Traditional chinese)
  eCharMapWansung,         ///< Wansung (Korean)
  eCharMapJohab,           ///< Johab (Korean)
  eCharMapAdobeStandard,   ///< Adobe Standard
  eCharMapAdobeExpert,     ///< Adobe Expert
  eCharMapAdobeCustom,     ///< Adobe Custom
  eCharMapAppleRoman,      ///< Apple Roman

  eCharMapFirst = eCharMapSymbol,     ///< First entry in charmap enum
  eCharMapLast = eCharMapAppleRoman   ///< Last entry in charmap enum
};


//! Font information
/*!
If the font has a fixed-size (ie. 'bitmap' type), the only valid global metrics
are the bitmap size (see nglFont1::GetSize). All metrics fields (from the bounding
box to the underline position) have no meaning and should not be used.

If the font is scalable (ie. 'vectorial' type), global metrics are given using
the current pixel size of the font. Hence these metrics depend on the last
nglFont1::SetSize call or the default font size applied at loading time.

Note on global metrics : (explane they are only hints used by the font
designer, show relation to glyph metrics)
*/
typedef struct nglFontInfo1
{
  nglString FamilyName;   ///< Family name (Helvetica, Times, Courier, etc)
  nglString StyleName;    ///< Style name (Italic, Bold, Stroke, etc)
  bool IsScalable;        ///< If true, font is freely scalable. If false only bitmap data is available.
  bool IsBold;            ///< True if the font is tagged as 'bold' (typographer hint)
  bool IsItalic;          ///< True if the font is tagged as 'italic' (typographer hint)
  uint GlyphCnt;          ///< Total number of available glyphs

  /** @name Global metrics (only relevant for scalable fonts)
    All metrics are given in pixels, they depend on the current font size
    (see nglFont1::SetSize)
  */
  //@{
  float BBoxMinX;         ///< Global bounding box, lower X coordinate
  float BBoxMinY;         ///< Global bounding box, lower Y coordinate
  float BBoxMaxX;         ///< Global bounding box, upper X coordinate
  float BBoxMaxY;         ///< Global bounding box, upper Y coordinate
  float Ascender;         ///< Ascender
  float Descender;        ///< Descender
  float Height;           ///< Height
  float AdvanceMaxW;      ///< Maximum horizontal advance
  float AdvanceMaxH;      ///< Maximum vertical advance
  float UnderlinePos;     ///< Underline Y coordinate
  float UnderlineThick;   ///< Underline thickness
  //@}
} nglFontInfo1;


//! Glyph metrics
/*!
All metrics are given in pixels, they depend on the current font size.
Note that vertical writing info (bearing and advance) is not included yet.
*/
typedef struct nglGlyphInfo1
{
  float Width;       ///< Bounding box width
  float Height;      ///< Bounding box height
  float HBearingX;   ///< X bearing (horizontal writing)
  float HBearingY;   ///< Y bearing (vertical writing)
  float HAdvance;    ///< Horizontal advance
} nglGlyphInfo1;


class _nglFontGlyph;
class _nglFontCache;


//! Font class
/*!
DEPRECATED, see nglFontBase
*/
class NGL_API nglFont1 : public nglError
{
public:
  typedef bool (*GlyphHandler)(FT_Bitmap* pBitmap, float Left, float Top, void* pUser);

  /** @name Life cycle */
  //@{
  nglFont1 (const nglPath& rFontFile, uint Face = 0);
  nglFont1 (const nglFont1& rFont);
  virtual ~nglFont1();
  //@}

  /** @name Font info & metrics */
  //@{
  bool GetInfo  (nglFontInfo1& rInfo);
  void DumpInfo (int Level);
  bool GetGlyphInfo  (nglGlyphInfo1& rInfo, uint Char);
  void DumpGlyphInfo (uint Char, int Level);
  bool IsScalable();

  float GetWidth();
  float GetHeight();
  bool  GetSize (float* pWidth, float* pHeight);
  bool  SetSize (float Width, float Height);
  void  UseKerning (bool Use = true);
  //@}

  /** @name Character maps (encoding) */
  //@{
  nglCharMap GetCharMap();
  bool       SetCharMap (nglCharMap Map);
  bool       HasCharMap (nglCharMap Map);
  //@}

  /** @name Rendering */
  //@{
  void UseGLConvention (bool Use = true);
  void UseAlpha     (bool Use = true);
  void UseAlphaTest (bool Use = true, GLenum Func = GL_GEQUAL, GLclampf Ref = 0.25f);
  void SetColor     (float R, float G, float B, float A = 1.0f);
  void SetGamma     (float Gamma);
  void SetPreserved (GLbitfield Mask = NGL_FONT1_PRESERVE_DEFAULT);

  int  Print (float X, float Y, const nglChar* pText, ...);
  int  Print (float X, float Y, const nglString& rText);
  int  GetTextSize (float& X, float& Y, const nglChar* pText); ///< Calculate the bounding of the string in texels and returns it in X & Y.
  int  GetTextPos (float X, const nglChar* pText); ///< Calculate the bounding of the char pos in the given text where the pixel (X,?) lies and returns it.

  FT_Glyph GetGlyph (uint Index) const;
  void SetGlyphHandler (GlyphHandler Handler, void* pUser);
  //@}

  /** @name Cache management */
  //@{
  bool AddCache (uint GlyphMin, const nglChar* pPrecached = NULL);
  //@}

private:
  static FT_Library     mFTLibrary;
  static FTC_Manager    mFTCacheManager;
  static FTC_ImageCache mFTImageCache;

  nglPath    mFacePath;
  uint       mFaceIndex;
  FT_Face    mFTFace;
  float      mSizeX;
  float      mSizeY;
  float      mGamma;
  uint8      mGammaRamp[256];
  float      mColor[4];
  bool       mUseKerning;
  bool       mCharMapConv;
  bool       mUseGLConv;
  GLbitfield mPreserved;
  bool       mUseAlpha;
  bool       mUseAlphaTest;
  GLenum     mAlphaTestFunc;
  GLclampf   mAlphaTestRef;
  std::list< _nglFontCache* > mCache;
  FTC_ImageTypeRec mFTImageType;
  GlyphHandler mGlyphHandler;
  void* mGlyphHandlerUser;

  void LoadFace (bool InitSize);
  int  GetGlyphIndexes (const nglChar* pSource, int SourceLength, FT_UInt32* pIndexes, int IndexesLength);
  bool GetGlyphIndexInfo (nglGlyphInfo1& rInfo, uint Index);
  bool GetGlyphTex  (FT_Bitmap* pBitmap, uint Width, uint Height, unsigned char*& pBuffer, bool& FreeBuffer);
  bool PrintGlyph   (uint Glyph, float& X, float& Y);
  bool PrintGlyphDef(uint Glyph, float& X, float& Y);

  float FUToPixelX (int FontUnit);
  float FUToPixelY (int FontUnit);
  float DotToPixelX (FT_F26Dot6 Dot);
  float DotToPixelY (FT_F26Dot6 Dot);

  FT_Encoding GetFTEncoding (nglCharMap Map);
  nglCharMap  GetNGLEncoding(FT_Encoding FTMap);

  GLsizei GetTex2DSize (uint Pixels, uint Max = 2048);
  uint    GetLog2Floor (uint Pixels, uint Max = 2^31);

  static FT_Error OnFaceRequest (FTC_FaceID FaceID, FT_Library pLibrary, FT_Pointer pData, FT_Face* pFace);
  static void Shutdown();

  virtual const nglChar* OnError (uint& rError) const;

  friend class _nglFontCache;
  friend class _nglFontGlyph;
};


/*
 * Internal classes
 */

#ifndef _DOXYGEN_
class _nglFontCache
{
private:
  _nglFontCache (nglFont1* pFont, uint GlyphMin, const nglChar* pPrecached);
  ~_nglFontCache();

  bool IsValid();
  bool PrintGlyph (uint Glyph, float& X, float& Y);

  nglFont1*       mpFont;
  GLuint          mTexName;             // GL texture object name
  GLint           mTexW, mTexH;         // Texture size (texels)
  char*           mpTexBuffer;          // Texture buffer (in GL ready format)
  _nglFontGlyph*  mpGlyph;              // Cached glyphs array
  std::map< uint, _nglFontGlyph* > mGlyphLUT; // mGlyph[GlyphIndex] -> _nglFontGlyph* CachedGlyph
  uint            mGlyphCnt, mGlyphMax;
  uint            mGlyphW, mGlyphH;
  GLfloat         mGlyphU, mGlyphV;

  friend class nglFont1;
  friend class _nglFontGlyph;
};


class _nglFontGlyph
{
private:
  _nglFontGlyph (nglFont1* pFont, _nglFontCache* pCache, uint Glyph, uint X, uint Y);

  bool IsValid();

  uint    mWidth, mHeight;     // Bitmap size (pixels)
  GLfloat mX1, mY1, mX2, mY2;  // Quad mapping coordinates
  float   mDX, mDY;            // Top & Left (from origin)
  float   mAdvX, mAdvY;        // Horizontal advance
  uint    mUseCnt;

  friend class nglFont1;
  friend class _nglFontCache;
};
#endif // _DOXYGEN_

#endif // __nglFont1_h__
