/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglFontLayout.h
\brief Text layouts (see nglFontBase)
*/

#ifndef __nglFontLayout_h__
#define __nglFontLayout_h__

#include "ngl.h"
#include "nglFontBase.h"


//! Describes a glyph item positioned in a layout
class NGL_API nglGlyphLayout
{
public:
  float X;      ///< Abscissa in layout coordinate space
  float Y;      ///< Ordinate in layout coordinate space
  int   Pos;    ///< Position of the source char in the original laid out string
  uint  Index;  ///< Glyph index in the current font
};


//! Layout context (see nglFontBase)
/*!
A layout context manages the positions of a glyph collection. When Layout() is
invoked, the OnGlyph() callback is triggered for every glyph extracted from the
original user string (a plain text). This base class does a very simple layout
job as a default.

To implement a more sophisticated layout, create your own layout class as a
nglFontLayout superclass, and override the OnGlyph() method. See OnGlyph()
documentation for more info.
*/
class NGL_API nglFontLayout
{
public:
  /** @name Life cycle */
  //@{
  nglFontLayout (const nglFontBase& rFont, float PenX = 0.0f, float PenY = 0.0f);
  /*!< Initialize a layout context
    \param rFont source font
    \param PenX initial horizontal pen position
    \param PenY initial vertical pen position

    Kerning is turned on as a default, see UseKerning().
  */
  virtual ~nglFontLayout();

  void Init (float PenX = 0.0f, float PenY = 0.0f);  ///< (re)initialize context
  //@}

  /** @name Metrics */
  //@{
  void UseKerning (bool Use = true);
  /*< Use kerning if available (turned on as a default)
   */
  int  GetMetrics (nglGlyphInfo& rInfo) const;
  /*!< Returns whole layout metrics (as a composite glyph)
    \param rInfo metrics info holder
    \return number of glyphs in the layout
  */
  //@}

  /** @name Layout */
  //@{
  int          Layout (const nglString& rText);
  /*!< 
    \param rText string to decompose
    \return number of glyphs effectively processed

    Extract a glyph collection from \p rText and build the layout. See OnGlyph().
  */
  virtual void OnGlyph (const nglString& rString, int Pos, nglGlyphInfo* pGlyph);
  /*!< Add a glyph to the layout
    \param rString current laid out string
    \param Pos position of the current char in the original laid out string
    \param pGlyph glyph object, might be null if the current char maps to no glyph

    This method inserts the given glyph into the current layout. It is a simple
    implementation which supports kerning and newline (\\n). Your own layout can
    be implemented by overloading this method. Here is a trivial example :

\code
void nglFontLayout::Onglyph (const nglString& rString, int Pos, nglGlyphInfo* pGlyph)
{
  if (!pGlyph) // No glyph available for this Char
    return;

  // Add this glyph to the layout with the current position pen position
  AddGlyph(mPenX, mPenY, Pos, pGlyph);

  // Proceed with glyph advance
  mPenX += pGlyph->AdvanceX;
  mPenY += pGlyph->AdvanceY;
}
\endcode

    Have a look at nglFontLayout::OnGlyph() default implementation for a little
    more involved exemple (newline and kerning support).

    It might be possible that a single char decomposes to several glyphs, in
    this case OnGlyph() will be called once per output glyph, with the same
    \p Pos parameter.
  */

  uint  GetGlyphCount() const;
  const nglGlyphLayout* GetGlyph   (uint Offset) const;
  const nglGlyphLayout* GetGlyphAt (float X, float Y) const;
  /*!< Identify a glyph at given coordinates
    \param X abscissa in layout coordinate space
    \param Y ordinate in layout coordinate space
    \return located glyph instance, NULL if none match

    This method searches if a glyph's bounding box in the layout encloses the
    given coordinates. If such a glyph is found, its localized instance is
    returned.
  */
  //@}

protected:
  typedef std::vector<nglGlyphLayout> GlyphList;

  const nglFontBase& mFont;       ///< Font source for glyphs
  float              mPenX;        ///< Current horizontal pen position
  float              mPenY;        ///< Current vertical pen position
  bool               mUseKerning;  ///< Use kerning in default layout
  uint               mGlyphPrev;   ///< Last laid out glyph (for kerning), 0 if none
  GlyphList          mGlyphs;      ///< List of glyphs

  bool AddGlyph   (float X, float Y, int Pos, nglGlyphInfo* pGlyph);
  /*!< Add a localized glyph to the layout
    \param X horizontal pen position
    \param Y vertical pen position
    \param Pos glyph position in original laid out string
    \param pGlyph glyph to insert, no insertion if NULL
    \return true if the glyph was succesfully appended to the layout

    Append a glyph to the current layout with pen position (\p X, \p Y).
    A nglGlyphLayout instance is added at the end of the mGlyphs vector.
    Internally, the global layout bounding box is updated.

    The \p Pos parameter is used for reverse mapping in GetGlyphAt(),
    to recover from a glyph back to the source character in the original
    laid out string. Multiple glyphs can reverse map to the same character
    index.
  */
  bool GetKerning (uint Index, float& rX, float& rY);
  /*!< Fetch kerning with previous glyph
    \param Index current glyph (right glyph of the kerning pair)
    \param rX horizontal kerning value
    \param rY vertical kerning value
    \return true if both \p rX and \p rY are valid

    The left glyph is defined by mGlyphPrev and only valid when non zero
    (thus the OnGlyph() might prevent kerning with the next glyph by resetting
    mGlyphPrev).
  */

private:
  float mXMin, mXMax;
  float mYMin, mYMax;

  void InitMetrics();
  void CallOnGlyph (const nglString& rString, int Pos, nglGlyphInfo* pGlyph);
};


#endif // __nglFontLayout_h__
