/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglIStream.h
\brief Input data stream base class
*/

#ifndef __nglIStream_h__
#define __nglIStream_h__

#include "ngl.h"
#include "nglString.h"
#include "nglStream.h"


//! Base class for input streams
/*!
This is a pure virtual class, all object capable of providing data should
heritate from nglIStream (see nglIFile for instance).
*/
class NGL_API nglIStream : public nglStream
{
public:
  /** @name Stream navigation */
  //@{
  virtual nglSize Available (uint WordSize = 1) = 0;
  /*!< Return the number of words immediatly available with any input method (or Peek)
    \param WordSize word size in bytes
    \return number of available words of \a WordSize bytes
  */
  //@}

  /** @name Input methods */
  //@{
  virtual nglSize ReadInt8  (uint8*  pData, nglSize Count = 1);  ///< Read \a Count 8-bit integer values, returns number of bytes effectively read
  virtual nglSize ReadInt16 (uint16* pData, nglSize Count = 1);  ///< Read \a Count 16-bit integer values, returns number of 16-bit words effectively read
  virtual nglSize ReadInt32 (uint32* pData, nglSize Count = 1);  ///< Read \a Count 32-bit integer values, returns number of 32-bit words effectively read
  virtual nglSize ReadInt64 (uint64* pData, nglSize Count = 1);  ///< Read \a Count 64-bit integer values, returns number of 64-bit words effectively read
  virtual nglSize ReadFloat (float*  pData, nglSize Count = 1);  ///< Read \a Count float values (IEEE 754 simple precision, 32 bits), returns number of 32-bit words effectively read
  virtual nglSize ReadDouble(double* pData, nglSize Count = 1);  ///< Read \a Count double values (IEEE 754 double precision, 64 bits), returns number of 64-bit words effectively read
  virtual nglSize ReadLine  (nglString& rLine, nglTextFormat* pFormat = NULL);
  /*!< Read a text of line
    \param rLine string object where the text line will be stored
    \param pFormat used to return the encountered line ending convention (if non-NULL)
    \return number of \e bytes read from stream

    This method attempts to read a line of text from the stream.
    The ending convention (see nglTextFormat) is automatically detected and returned
    via *\a pFormat (if possible). If the end of stream was reached before the invocation
    or an error occured, the method reports zero bytes read and *\a pFormat is left
    untouched. \b Note : SetTextFormat() and GetTextFormat() only set the line-ending
    convention for output, they have no relation with this method.

    Text encoding conversion is performed if necessary with nglStringImport defaults.
    See also SetTextEncoding() and GetTextEncoding().

    \a rLine is always emptied (even if an error occured), and then filled with the
    encountered text line. This string does contain the \\n line ending character,
    unless it was the last line of the stream and it didn't have a line ending (in
    this case \p *pFormat will also be set to \e eTextNone).

    \b Example : parse a stream line by line, display each of them with a number

\code
void DumpLines(nglIStream& rInput)
{
  nglString line;
  uint count = 0;

  while (rInput.ReadLine(line))
    NGL_OUT (_T("%d: %s"), ++count, line.GetChars());

  NGL_OUT(_T("Total: %d line%s\n"), count, PLURAL(count));
}
\endcode
  */
  virtual nglSize ReadText  (nglString& rText, nglTextFormat* pFormat = NULL);
  /*!< Read a whole text file at once
    \param rText string object where the text will be stored
    \param pFormat used to return the encountered line ending convention (if non-NULL).
                   If several line endings are encountered during parsing, the best
                   occurence is reported.
    \return number of \e bytes read from stream

    This method is equivalent to the following program (but more optimized, and with
    line ending weighting) :
\code
nglSize ReadText (nglString& rText, nglTextFormat* pFormat);
{
  nglString line;
  nglSize done = 0;

  while ((done =+ ReadLine(line)))
    rText += line;

  return done;
}
\endcode
  */
  virtual nglSize Read (void* pData, nglSize WordCount, uint WordSize) = 0;
  /*!< Read \a WordCount number of \a WordSize bytes long into \a pBuffer
    \param pData a buffer which may hold \a WordCount words of \a WordSize bytes
    \param WordCount number of words to read
    \param WordSize word size in bytes
    \return number of words effectively read

    Attempts to read data by \e word units of \a WordSize bytes. This method will not
    retrieve incomplete words (when there are not enough available bytes to make a word).
  */
  virtual nglSize Peek (void* pData, nglSize WordCount, uint WordSize);
  /*!< Peek \a WordCount number of \a WordSize bytes long into \a pBuffer
    \param pData a buffer which may hold \a WordCount words of \a WordSize bytes
    \param WordCount number of words to read
    \param WordSize word size in bytes
    \return number of words effectively read

    This method works as Read() but does not change the stream current position.
  */
  //@}

  /** @name Input methods (alternate) */
  //@{
  uint8   ReadInt8   (uint8 Default = 0);     ///< Read an 8-bit integer value, returns \a Default if no data is available
  uint16  ReadInt16  (uint16 Default = 0);    ///< Read a 16-bit integer value, returns \a Default if no data is available
  uint32  ReadInt32  (uint32 Default = 0);    ///< Read a 32-bit integer value, returns \a Default if no data is available
  uint64  ReadInt64  (uint64 Default = 0);    ///< Read a 64-bit integer value, returns \a Default if no data is available
  float   ReadFloat  (float Default = 0.0);   ///< Read a float value (IEEE 754 simple precision, 32 bits), returns \a Default if no data is available
  double  ReadDouble (double Default = 0.0);  ///< Read a double value (IEEE 754 double precision, 64 bits), returns \a Default if no data is available
  //@}

private:
  bool GetConv();
};

#endif // __nglIStream_h__
