/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglImageCodec.h
\brief Base class for nglImage pluggable loader/exporter
*/

#ifndef __nglImageCodec__
#define __nglImageCodec__

#include "ngl.h"
#include "nglError.h"
#include "nglIStream.h"
#include "nglOStream.h"
#include "nglPath.h"
#include "nglImage.h"


//! Image codec description
/*!
Every image codec (loader and/or writer) registers itself to nglImage
passing its description via this class.
*/
class NGL_API nglImageCodecInfo
{
public:
  nglImageCodecInfo()
  {
  }

  virtual ~nglImageCodecInfo()
  {
  }

  virtual bool ExtensionMatch (nglString& rFileName)
  {
    std::vector<nglString>::iterator ext;
    int filename_len = rFileName.GetLength();

    for (ext = mExtensions.begin(); ext < mExtensions.end(); ext++)
    {
      int ext_size = (*ext).GetLength();

      if (!rFileName.Compare (*ext, filename_len - ext_size, ext_size, false))
        return true;
    }
    return false;
  }

  virtual nglImageCodec* CreateInstance() = 0; ///< Creates an instance of the corresponding codec

  bool mCanSave;                  ///< True if a codec can encode to a byte stream
  bool mCanLoad;                  ///< True if the codec can decode a byte stream
  nglString mName;                ///< The name of the codec (ie. 'JPEG', 'PNG', 'TGA', etc)
  std::vector<nglString> mExtensions;  ///< List of file name extensions for this kind of data
  nglString mInfo;                ///< Whatever ! (copyright info, user manual, etc)
};


//! Base class for nglImage pluggable loader/exporter
/*!
FIXME
*/
class NGL_API nglImageCodec 
{
public:
  virtual ~nglImageCodec()
  {
  }
  virtual bool Init (nglImage* pImage)      ///< Codec init is decoupled from construction
  {
    mpImage = pImage;
    return true;
  }
  virtual bool Probe(nglIStream* pIStream) = 0;  ///< Check for a known signature at stream start. Returns true if it wants to take the job
  virtual bool Feed (nglIStream* pOStream) = 0; 
  virtual bool Save (nglOStream* pOStream) = 0; 
  virtual float GetCompletion() = 0; 

protected:
  nglImage* mpImage;

  void SendInfo (nglImageInfo& rInfo)  ///< Send image description to image object. The image object will allocate the image buffer.
  {
    if (mpImage) mpImage->OnCodecInfo (rInfo);
  }
  void SendData (float Completion)  ///< Acknowledge that more data was decoded to image buffer
  {
    if (mpImage) mpImage->OnCodecData (Completion);
  }
  bool SendError()  ///< Signals an encoding/decoding error
  {
    return (mpImage) ? mpImage->OnCodecError() : false;
  }
};

#endif // __nglImageCodec_h__
