/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglKernel.h
\brief Kernel, application abstraction base class
*/

#ifndef __nglKernel_h__
#define __nglKernel_h__

#include "ngl.h"
#include "nglEvent.h"
#include "nglError.h"
#include "nglString.h"
#include "nglLog.h"
#include "nglConsole.h"
#include "nglPath.h"
#include "nglTimer.h"
#include <list>

#ifdef _X11_
#include <X11/Xlib.h>
#endif // _X11_


//! Kernel, application abstraction base class
/*!
NGL is a framework which can run in various environments. The most trivial one
is a regular application. A more advanced scenario would be running as a
plugin into a hosting application, such as a web browser for instance.

The kernel is the common interface to all these environments : wether your
program is compiled as a regular application or a plugin, one and only one
instance of nglKernel is always available in a single environment (to be
more precise, in a single process). It provides these essential services :

- initialisation and exit callbacks

- event manager interfaces, used internally by timers, windows, etc

- a global and unique <em>application log</em>, which is used by almost any
  instance of a regular NGL class

- a global and unique <em>application console</em>, which is the default
  output for the <em>application log</em>. See nglConsole for more info
  on the default console

- various other global services

nglKernel is an abstract class. You must refer to nglKernel derivates to
create an instance. See the ubiquitous nglApplication for instance.

The unique nglKernel object is always available via the \a App pointer.
*/
class NGL_API nglKernel : public nglError, public nglEvent
{
public:
  typedef void (*ExitFunc) (void); ///< Cleanup callback type. See AddExit()

  /** @name Life cycle */
  //@{
  virtual void Quit(int Code = 0) = 0;
  /*!<
    Request application ending
    \param Code exit code

    Note that some instances (such as plugins) might ignore a call to this method.
  */
  virtual void AddExit (ExitFunc Func);
  /*!< Register a shutdown callback
    \param Func callback function, must be either a C function or a static C++ method

    Register a callback to be triggered when the application is shuting down.
    Handlers are called in reverse order of registering (last registered is called first).
  */
  virtual void DelExit (ExitFunc Func);
  /*!< Unregister a shutdown callback
    \param Func callback function

    Unregister a callback registered with AddExit().
  */
  //@}

  /** @name Runtime debugging */
  //@{
  virtual void SetDebug(bool Debug);
  /*!<
    Set run-time debug status

    Code enclosed in the DEBUG() macro can be (des)activated at run-time either with SetDebug(),
    either with the '--nodebug' command-line option if you invoke nglApplication::ParseDefaultArgs().
  */
  virtual bool GetDebug() { return mDebug; }
  /*!<
    Return run-time debug status

    See SetDebug().
  */
  //@}

  /** @name Logging and output */
  //@{
  virtual nglLog& GetLog();
  /*!<
    Gives the application log.

    Notably used by the NGL_OUT() and NGL_LOG() macros. The available log domains
    are defined in ngl.h, here is the list :

    - Default : if you not specify a domain, application logs are directed there.
    - NGL_DOMAIN_* : please consult the bottom of ngl.h for available log domains
  */
  virtual nglConsole& GetConsole();
  /*!<
    Gives the application console.
  */
  virtual void SetConsole(nglConsole* pConsole);
  /*!<
    Set application console. User still owns the \p pConsole objects
    (ie. remember to delete it on application exit).
  */
  //@}

  /** @name Application context */
  //@{
  nglPath GetPath();
  /*!<
    \return current application's path (includes the binary file name)
  */
  nglString GetName();
  /*!<
    \return current application name

    The application name is implicitely obtained from the application's executable
    name. However it can be set explicitely with SetName() at any time. The application
    name serves at least two purposes :

    - allow per-application resources settings (TODO: resources!)
    - nglWindow objects will use the application name as their default title
  */
  int       GetArgCount();
  /*!<
    Return the number of user arguments passed along the execution of the
    program. \b Caution \b ! The executable name (or invocation name) is
    not included, use GetName() for this purpose.
  */
  nglString GetArg(int Index);
  /*!<
    Return one of the user arguments passed along the execution of the
    program. \a Index must be between 0 (first argument) and GetArgCount() - 1.
    \b Caution \b ! The executable name (or invocation name) is
    not included, use GetName() for this purpose.
    If \a Index is out of range, returns an empty string.
  */
  //@}

  /** @name Clipboard */
  //@{
  virtual nglString GetClipboard();
  /*!<
    Get the clipboard text content
    \return If there is text in the clipboard, returns it, or return an empty string
  */
  virtual bool SetClipboard(const nglString& rString);
  /*!<
    Set the text content of the clipboard
    \param rString the text to put into the clipboard
    \return returns false if the clipboard is not available (maybe temporarily)
  */
  //@}

  /** @name User callbacks */
  //@{
  virtual void OnInit() = 0;
  /*!<
    This method is called at startup time, as soon as the application is ready to run.

    There is always at least one argument, (ArgCnt > 0) and it contains the application
    name. User parameters are not available on all platforms (MacOS being one
    of them).
  */
  virtual void OnExit (int Code);
  /*!<
    This method is called when the application exits.
    \param Code exit code

    The exit code is returned by the process to the system. It should be 0
    for normal termination (ie. no error encountered), positive otherwise.

    The application only quits if :
    - Quit() has been explicitely called and the control has been given back to the event loop
    - the process received a signal (\e Unix \e only). HUP, INT, QUIT, PIPE and TERM are
      currently intercepted.
  */
  //@}

protected:
  // Life cycle
  nglKernel();
  virtual ~nglKernel();
  void IncRef();
  void DecRef();
  void Init();
  void Exit();

  void SetName (const nglString& rName);
  void SetPath (const nglPath& rPath);
  void ParseCmdLine (char* pCmdLine);
  void AddArg (const nglString& rArg);
  void DelArg (int Pos, int Count = 1);

  // Internal callbacks
  void CallOnInit();
  void CallOnExit(int Code);

  // From nglError
  virtual const nglChar* OnError (uint& rError) const = 0;

private:
  uint                   mRefCount;
  bool                   mDebug;
  nglLog*                mpLog;
  nglConsole*            mpCon;
  bool                   mOwnCon;
  std::list<ExitFunc>    mExitFuncs;
  nglPath                mPath;
  nglString              mName;
  std::vector<nglString> mArgs;

//  virtual void OnEvent (uint Flags);  // From nglEvent

#ifdef _WIN32_
protected:
  friend class nglConsole;
  friend class nglTimer;
  friend class nglWindow;

  virtual HINSTANCE GetHInstance();
  virtual HWND      GetHWnd();
  virtual HDC       GetDC();
  virtual HFONT     GetSystemFont();

  bool SysInit(HINSTANCE Inst);

private:
  HINSTANCE  mHInstance;
  HDC        mHDC;
  HWND       mHWnd;
  HFONT      mSystemFont;
  int        mResolutionChanged;
  bool       mShowConsole;

  static LRESULT CALLBACK MainWndProcStatic (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
  LRESULT                 MainWndProc       (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);

  friend class nglPluginKernel;
  friend class nglPlugin;
  BOOL DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved);
#endif // _WIN32_

#ifdef _UNIX_
public:
  virtual void     AddEvent (nglEvent* pEvent) = 0;
  virtual void     DelEvent (nglEvent* pEvent) {};
  virtual void     AddTimer (nglTimer* pTimer) = 0;
  virtual void     DelTimer (nglTimer* pTimer) {};
#ifdef _X11_
  virtual Display* GetDisplay() = 0;
  virtual void     AddWindow (class nglWindow* pWin) = 0;
  virtual void     DelWindow (class nglWindow* pWin) {};
#endif // _X11_

protected:
  bool             SysInit();
  virtual void     OnEvent(uint Flags); // From nglEvent

private:
  void             CatchSignal (int Signal, void (*pHandler)(int));
  static void      OnSignal (int Signal);
#endif // _UNIX_

#ifdef _CARBON_
protected:
  bool             SysInit();
  bool mExitPosted;
#endif // _CARBON_
};


extern NGL_API nglKernel* App;
/*!<
A well designed program should have only one global. This is the unique NGL global,
pointing to the unique instance of the kernel. Since nglKernel is an abstract class,
\a App real type is necessary a descendant of nglKernel. See nglApplication and
the NGL_APP_CREATE macro for instance.
*/


/*
 * NGL kernel log/debug macros
 */

// NGL_OUT("i=%d\n", i)
#define NGL_OUT  App->GetConsole().Output
#define NGL_OUTV App->GetConsole().Outputv

// NGL_LOG(5, "i=%d\n", i)
// NGL_LOG(5, 2, "i=%d\n", i)
// NGL_LOG("net", "i=%d\n", i)
// NGL_LOG("net", 2, "i=%d\n", i)
#define NGL_LOG  App->GetLog().Log
#define NGL_LOGV App->GetLog().Logv

/* Let's use Win32 assert() when available (pops up a dialog in debug mode)
 */
#ifdef _WIN32_
#  include <assert.h>
#  define __ASSERT_SYS(test) assert(test);
#else
#  define __ASSERT_SYS(test)
#endif

#define __T(x) _T(x)
#define __TFILE__ __T(__FILE__)

#define __NGL_ASSERT(test) \
  if (!(test)) { NGL_OUT(_T("%s(%d): " #test " failed\n"), __TFILE__, __LINE__); __ASSERT_SYS(test) }

#define __NGL_ASSERTR(test, result) \
  if (!(test)) { NGL_OUT(_T("%s(%d): " #test " failed\n"), __TFILE__, __LINE__); __ASSERT_SYS(test) return result; }

#ifdef _DEBUG_
#  define NGL_ASSERT(x)    __NGL_ASSERT(x)
#  define NGL_ASSERTR(x,y) __NGL_ASSERTR(x,y)
#else
#  define NGL_ASSERT(x)
#  define NGL_ASSERTR(x,y)
#endif // _DEBUG_

/* DEBUG facility
 */
#ifdef _DEBUG_
#  define NGL_DEBUG(x) if (App->GetDebug()) { x }
#else
#  define NGL_DEBUG(x)
#endif

#endif // __nglKernel_h__
