/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglLog.h
\brief Log and debug facilities
*/

#ifndef __nglLog_h__
#define __nglLog_h__

#include "ngl.h"
#include "nglError.h"
#include "nglString.h"
#include "nglOStream.h"
#include <stdarg.h>
#include <vector>


//! Logging domain structure
/*!
This structure helps the user to statically define a logging domain. Example :

\code
#define DOMAIN_A 1
#define DOMAIN_B 2

nglLogDomainInfo* DomainA = {Index: DOMAIN_A; Level: 2; pAlias: "A"; pName: "Domain A"};
nglLogDomainInfo* DomainB = {DOMAIN_B, 5, "B", "Domain B"};
\endcode

The \a pName log name is prepended to all log output using the corresponding
domain. Log domains are integer for the programmer, and is set by the
\a Index value. It is recommended to use macros to hide domain index
values. The \a Level parameter sets the initial verbose level, it can
be changed later with the nglLog::SetLevel() methods.

While it is possible to use log domains only with their names (without
using any index at all), it is not recommended.
*/
class nglLogDomainInfo
{
public:
  uint           Index;   ///< Domain index (starting from 0, the default domain)
  int            Level;   ///< Initial verbose level
  const nglChar* pAlias;  ///< Domain alias
  const nglChar* pName;   ///< Domain name (for display purposes)
};


//! Logging domain helper for nglLog
/*!
This is a simple holder for a logging domain definition. This class is
currently opaque to the user, and only used internally by nglLog.
*/
class NGL_API nglLogDomain
{
private:
  int       mLevel;
  nglString mAlias;
  nglString mName;

  nglLogDomain(int Level, const nglChar* pAlias, const nglChar* pName);

friend class nglLog;
};

//! Logging flags
/*!
These flags are currently used to control the tags automatically added at
the beginning of every logged line. Availables values are :

- nglLog::NoStamp
- nglLog::TimeStamp
- nglLog::DateStamp
- nglLog::DomainStamp (default)
*/
typedef uint nglLogFlag;


//! Logging facilities
/*!
The nglLog class is a sophisticated gadget to manage event logging or simple
debugging. A nglLog object can have :
- multiple outputs, such as console, file or network
- many domains, each with their own verbose level
- configurable stamping of events
- uniform output in any case, easily parsable for diagnosis

Domains let you easily classify the events in your application : if you follow
some simple guidelines, you will be able to tune very precisely which messages
you want to log and their given verbose level.

NGL programs have an <em>application log</em> which is always created at
startup and used internally by NGL to report errors and other useful
informations. Have a look at nglApplication::ParseDefaultArgs() to see how this
facility can help you.

The easiest method to set up your own nglLog object populated with domains is
called the 'static method'. First, define your logging domains in a common
include file for your project :

\code
// File TestApp.h
#include "nglApplication.h"

#define DOMAIN_NET   (uint)1
#define DOMAIN_GFX   (uint)2
#define DOMAIN_MISC  (uint)3

class TestApp : public NGL_APP_TYPE
{
public:
  OnInit();
  DoIt();

private:
  nglLog* mpTestLog;
}
\endcode

Use a consecutive numbering starting frome one (this is not mandatory, but only
a good piece of advice). Then you define a static logging domain table and use it
in your constructor :

\code
// File TestApp.cpp
#include "TestApplication.h"

/\* Column 1 : domain index
 * Column 2 : initial verbose level
 * Column 3 : domain alias
 * Column 4 : domain display name
 *\/
const nglLogDomainInfo TestAppDomains[] = {
  {0,           1, "default", ""},     // '0' is the default domain
  {DOMAIN_NET,  2, "net",     "Net "},
  {DOMAIN_GFX,  1, "gfx",     "Gfx "},
  {DOMAIN_MISC, 0, "misc",    "Misc"},
  {NULL, 0, NULL}
};

TestApp::OnInit()
{
  mpTestLog = new nglLog (true); // Log using the console
  mpTestLog->AddDomains (TestAppDomains);
  DoIt();
}

TestApp::DoIt()
{
  mpTestLog->Log (DOMAIN_MISC, 0, "Here is an unmaskable event from the misc domain");
  mpTestLog->Log (DOMAIN_MISC, "This is strictly equivalent to the preceding line");
  mpTestLog->Log (DOMAIN_GFX, 3, "This one will only be displayed if gfx's verbose level is 3 or more");
  mpTestLog->SetLevel (DOMAIN_NET, 5);
  mpTestLog->Log (DOMAIN_NET, 2, "This one will be displayed, since 2 <= 5");
}
\endcode

The verbose level rule is simple : an event is logged only if the level passed
to an output method is equal or less than the current domain verbose level.
For instance in our example, the initial verbose level of the DOMAIN_MISC domain
is set to 0 : hence only events using a 0 level (or lower, but using negative log
level is not a good practice) will be displayed.

This is the reason why events using the 0 level are called 'unmaskable events',
since they are displayed even if the domain verbose level is set to 0.

There is no general rule for level assignation, howevere here are some good tips :
- Use level 0 only for important messages, the kind everybody (even end users)
  should see.
- Use level 1-7 for basic and low frequency information, such as startup and
  shutdown details.
- Use level 8-63 for repeating events, such as graphic events (mouse clic,
  key press and so on). Keep messages well formated, clear and easy.
- User level 64-... for high frequency or internal events only. This information
  often only makes sense for the code owner.
*/

class NGL_API nglLog : public nglError
{
public:
  static const nglLogFlag NoStamp;      ///< Do not stamp log lines (raw output)
  static const nglLogFlag TimeStamp;    ///< Add a time stamp (HH:MM:SS format)
  static const nglLogFlag DateStamp;    ///< Add a date stamp (YY/MM/DD format)
  static const nglLogFlag DomainStamp;  ///< Add a domain stamp ([domain])

  /** @name Life cycle */
  //@{
  nglLog(bool UseConsole = true);
  /*!<
    (Default) constructor
    \param UseConsole output events to the console

    You can add outputs later. Set \a UseConsole explicitly to false if you
    want to use only alternate methods (file, network). See the AddOutput()
    and UseConsole() methods.
  */
  virtual ~nglLog();
  /*!< Flush all pending events and clos all current outputs. */
  //@}

  /** @name Formatting flags */
  //@{
  void       SetFlags (nglLogFlag Flags);  ///< Set all flags at once
  void       SetFlag  (nglLogFlag Flag);   ///< Set a flag
  void       UnsetFlag(nglLogFlag Flag);   ///< Unset a flag
  nglLogFlag GetFlags ();                  ///< Get current flags
  //@}

  /** @name Output selection */
  //@{
  void       UseConsole(bool Use);  ///< Send log output to the application's console
  bool       AddOutput (nglOStream* pStream);
  /*!< Add an output stream
    \param pStream new output stream
    \return true if succeeded
  */
  bool       DelOutput (nglOStream* pStream);
  /*!< Remove an output stream
    \param pStream output stream to remove
    \return true if succeeded
  */
  //@}

  /** @name Domain management */
  //@{
  bool       AddDomain (uint Index, int Level, const nglChar* pAlias, const nglChar* pName);
  /*!<
    Add a logging domain.
    \param Index domain index
    \param Level initial verbose level
    \param pAlias domain alias
    \param pName domain display name

    This function will fail, return false and set the NGL_LOG_EUSED error when
    the domain index \a Index is already used.
  */
  bool       AddDomains(const nglLogDomainInfo pDomains[], int Count = -1);
  /*!<
    Add an array of logging domains.
    \param pDomains array of nglLogDomainInfo domain definitions
    \param Count number of element in array

    If the array is NULL-terminated, the \a Count parameter can be safely
    ommited or set to -1.

    This function will fail, return false and set the NGL_LOG_EUSED error when
    an index of at least one domain in the table is already used.
  */
  int        GetIndex  (const nglChar* pAlias);
  /*!<
    Get the index from a domain alias.
    \param pAlias domain alias

    Return a negative value if the domain alias \a rAlias is not found.
  */
  nglString  GetAlias   (uint Domain);
  /*!<
    Get the domain alias from an index.
    \param Domain domain index

    Return a NULL pointer if the index domain \a Domain is not found.
  */
  nglString  GetName   (uint Domain);
  /*!<
    Get the domain display name from an index.
    \param Domain domain index

    Return a NULL pointer if the index domain \a Domain is not found.
  */
  int        GetLevel  (uint Domain);
  /*!<
    Get the verbose level of a domain.
    \param Domain domain index

    Return a negative value if the index domain \a Domain is not found.
  */
  int        GetLevel  (const nglChar* pDomain);
  /*!<
    Get the verbose level from a domain name.
    \param pDomain domain name

    Return a negative value if the domain name \a rDomain is not found.
  */
  bool       SetLevel  (uint Domain, int Level);
  /*!<
    Set the verbose level of a domain.
    \param Domain domain index
    \param Level new verbose level

    Return false if the index domain \a Domain is not found.
  */
  bool       SetLevel  (const nglChar* pDomain, int Level);
  /*!<
    Set the verbose level of a domain.
    \param pDomain domain name
    \param Level new verbose level

    Return false if the domain name \a rDomain is not found.
  */
  //@}

  /** @name Output methods */
  //@{
  void       Output    (const nglChar* pText, ...);
  /*!<
    Output event without verbose level test.
    \param pText format string
    \param ... arguments

    This method bypass the verbose level test, and does a raw output (no
    stamping at all). It lets you reach the selected outputs directly.
  */
  void       Log       (uint Domain, const nglChar* pText, ...);
  /*!<
    Log an unmaskable event (use level 0).
    \param Domain domain index
    \param pText format string
    \param ... arguments
  */
  void       Log       (uint Domain, int Level, const nglChar* pText, ...);
  /*!<
    Log an event.
    \param Domain domain index
    \param Level verbose level
    \param pText format string
    \param ... arguments
  */
  void       Log       (const nglChar* pDomain, const nglChar* pText, ...);
  /*!<
    Log an unmaskable event (use level 0).
    \param pDomain domain name
    \param pText format string
    \param ... arguments
  */
  void       Log       (const nglChar* pDomain, int Level, const nglChar* pText, ...);
  /*!<
    Log an event.
    \param pDomain domain name
    \param Level verbose level
    \param pText format string
    \param ... arguments
  */
  void       Logv      (uint Domain, const nglChar* pText, va_list Args);
  void       Logv      (uint Domain, int Level, const nglChar* pText, va_list Args);
  void       Logv      (const nglChar* pDomain, const nglChar* pText, va_list Args);
  void       Logv      (const nglChar* pDomain, int Level, const nglChar* pText, va_list Args);
  //@}

protected:
  virtual const nglChar* OnError (uint& rError) const;

private:
  bool                       mUseConsole;
  nglLogFlag                 mFlags;
  nglString                  mOutputBuffer;
  nglString                  mPrefix;
  nglString                  mBody;
  
  std::vector<nglOStream*>  mOutputs;
  std::vector<nglLogDomain*> mDomains;

  nglLogDomain* Find(uint Index);
  nglLogDomain* Find(const nglChar* pName);
  void          InternalLog (const nglLogDomain* pDomain, int Level, const nglChar* pText, va_list Args);
  void          Dispatch (const nglString& rText);
};

#endif // __nglLog_h__
