/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglMath.h
\brief Mathematical constants, functions and helpers
*/

#ifndef __nglMath_h__
#define __nglMath_h__

#include "ngl.h"
#include <math.h>
#include "nglEndian.h"


#if defined(_WIN32_) || defined(__MWERKS__)
  // These defines are not present in <math.h> with VC6 sp3
  #define M_E            2.7182818284590452354   /* e */
  #define M_LOG2E        1.4426950408889634074   /* log_2 e */
  #define M_LOG10E       0.43429448190325182765  /* log_10 e */
  #define M_LN2          0.69314718055994530942  /* log_e 2 */
  #define M_LN10         2.30258509299404568402  /* log_e 10 */
  #define M_PI           3.14159265358979323846  /* pi */
  #define M_PI_2         1.57079632679489661923  /* pi/2 */
  #define M_PI_4         0.78539816339744830962  /* pi/4 */
  #define M_1_PI         0.31830988618379067154  /* 1/pi */
  #define M_2_PI         0.63661977236758134308  /* 2/pi */
  #define M_2_SQRTPI     1.12837916709551257390  /* 2/sqrt(pi) */
  #define M_SQRT2        1.41421356237309504880  /* sqrt(2) */
  #define M_SQRT1_2      0.70710678118654752440  /* 1/sqrt(2) */
#endif // _WIN32 || __MWERKS__

#define DEG2RAD(x) ((x) * M_PI / 180.0f)
#define RAD2DEG(x) ((x) * 180.0f / M_PI)


uint NGL_API Log2Ceil(uint Value);
/*!< Retrieve the binary logarithm of \a Value (by excess)
  \return the smallest x such as \a Value <= 2^x

\code
Log2Ceil(30) : 5
Log2Ceil(40) : 6
Log2Ceil(64) : 6
Log2Ceil(65) : 7
\endcode
*/

uint NGL_API Log2Floor(uint Value);
/*!< Retrieve the binary logarithm of \a Value (by default)
  \return the smallest x such as \a 2^x <= Value < 2^(x+1)

\code
Log2Floor(30) : 4
Log2Floor(40) : 5
Log2Floor(64) : 6
Log2Floor(65) : 6
\endcode
*/


/* ISOC99 round() functions emulation when not supported
 */
#if defined(_WIN32_) || defined(_FREEBSD_)
  inline double round(double x)
  {
    double r = ceil(x);
    return r - x <= 0.5f ? r : r - 1.0f;
  }
  inline float roundf(float x)
  {
    float r = (float)ceil((float)x);
    return r - x <= 0.5f ? r : r - 1.0f;
  }
#endif // _WIN32_ || _FREEBSD_


/* Float to Ints Conversion helpers
 */
#if (BYTE_ORDER == BIG_ENDIAN)
  #define __iexp 0
  #define __iman 1
#else
  #define __iexp 1
  #define __iman 0
#endif

inline int32 ToBelow(double x) ///< Cast x to an int using "to zero" rounding mode (Normal Ansi C behaviour).
{
#ifdef USE_FAST_FLOAT_CONVERSION
  const double _double2fixmagic = 68719476736.0*1.5;     //2^36 * 1.5,  (52-_shiftamt=36) uses limited precisicion to floor
  const int32 _shiftamt         = 16;                    //16.16 fixed point representation,

  x = x + _double2fixmagic;
  return (((uint32*)&x)[__iman] >> _shiftamt)
       | (((uint32*)&x)[__iexp] << _shiftamt); 
#else
  return (int32) floor(x);
#endif
}

#undef __iexp
#undef __iman

inline int32 ToBelow(float x) ///< Cast x to an int using "to zero" rounding mode (Normal Ansi C behaviour).
{
	return ToBelow ((double)x);
}

inline int32 ToZero(float x)  ///< Cast x to an int using "to lower" rounding mode.
{
  return (x < 0.f ? 1 : 0) + ToBelow((double)x);
}

inline int32 ToZero(double x)  ///< Cast x to an int using "to lower" rounding mode.
{
  return (x < 0.f ? 1 : 0) + ToBelow(x);
}

inline int32 ToNearest(float x) ///< Cast x to an int using "to nearest" rounding mode.
{
  return ToZero(x + (x >= 0 ? .5f : - .5f));
}

inline int32 ToNearest(double x) ///< Cast x to an int using "to nearest" rounding mode.
{
  return ToZero(x + (x >= 0 ? .5 : - .5));
}

inline int32 ToAbove(float x) ///< Cast x to an int using "to greater" rounding mode.
{
  return -ToBelow((double)-x);
}

inline int32 ToAbove(double x) ///< Cast x to an int using "to greater" rounding mode.
{
  return -ToBelow(-x);
}

#endif // __nglMath_h__
