/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglMatrix.h
\brief Homogeneous matrixes template classes
*/

#ifndef __nglMatrix_h__
#define __nglMatrix_h__

#include "ngl.h"
#include "nglMath.h"
#include "nglVector.h"
#include "nglApplication.h"
#include <string.h>


//! Homogenous 4x4 matrixes
/*!
This class implements the usual operations on 4x4 homogenous matrixes used
in OpenGL, as a client side counterpart of the OpenGL matrix manipulations
functions. Remember that this implementation can't benefit of any hardware
accelerated TnL, and that you should only perform minimal geometry computing
without OpenGL.

The internal representation is the same as the OpenGL, it is ordered as
column first. Illustration :

\code
  // Using the Array field, we have this layout :
  //
  //   Array[0]  Array[4]  Array[8]  Array[12]
  //   Array[1]  Array[5]  Array[9]  Array[13]
  //   Array[2]  Array[6]  Array[10] Array[14]
  //   Array[3]  Array[7]  Array[11] Array[15]

  // Using the named elements from Elt instead of the array, we have this layout :
  //
  //   Elt.M11  Elt.M12  Elt.M13  Elt.M14
  //   Elt.M21  Elt.M22  Elt.M23  Elt.M24
  //   Elt.M31  Elt.M32  Elt.M33  Elt.M34
  //   Elt.M41  Elt.M42  Elt.M43  Elt.M44
\endcode

You have some helpers to integrate nglMatrix<> usage seamlessly with OpenGL :

\code
nglMatrix<float> mat;

mat.SetRotation(45, nglVectorf(1, 0, 0));
mat.LoadMatrix();
\endcode

... does the same job as :

\code
nglMatrix<float> mat;

mat.SetRotation(45, nglVectorf(1, 0, 0));
glLoadMatrixf(mat.Array);
\endcode

... and again the same thing :

\code
glLoadIdentity();
glRotatef(45, 1, 0, 0);
\endcode
*/
template <class T> class nglMatrix
{
public:
  union
  {
    T Array[16];
    struct 
    {
      T M11,M21,M31,M41;
      T M12,M22,M32,M42;
      T M13,M23,M33,M43;
      T M14,M24,M34,M44;
    } Elt;
  };

  /// Create an identity matrix (see SetIdentity())
  nglMatrix()
  {
    SetIdentity();
  }

  /*! Create a matrix from an array
    \param pArray flat array (see nglMatrix for conventions), nullify all elements if NULL
  */
  nglMatrix(const T pArray[16])
  {
    if (pArray)
      memcpy (Array, pArray, 16*sizeof(T));
    else
      Nullify();
  }

  nglMatrix(const nglMatrix<T>& rMatrix)
  {
    memcpy (Array, rMatrix.Array, 16*sizeof(T));
  }

  virtual ~nglMatrix()
  {
  }

  /// Set all elements to zero
  void Nullify()
  {
    uint i;
    for (i = 0; i < 16; i++)
      Array[i] = 0.f;
  }

  /// Set the matrix to identity
  void SetIdentity()
  {
    Nullify();
    Elt.M11 = 1.f;
    Elt.M22 = 1.f;
    Elt.M33 = 1.f;
    Elt.M44 = 1.f;
  }

  /// Transpose the matrix
  void Transpose ()
  {
    T x;

    x = Array[1];  Array[1]  = Array[4];  Array[4]  = x;
    x = Array[2];  Array[2]  = Array[8];  Array[8]  = x;
    x = Array[3];  Array[3]  = Array[12]; Array[12] = x;
    x = Array[6];  Array[6]  = Array[9];  Array[9]  = x;
    x = Array[7];  Array[7]  = Array[13]; Array[13] = x;
    x = Array[11]; Array[11] = Array[14]; Array[14] = x;
  }

  /// Compute the determinant
  T Determinant ()
  {
    return
       (Elt.M11*Elt.M22 - Elt.M12*Elt.M21)*(Elt.M33*Elt.M44 - Elt.M34*Elt.M43)
      -(Elt.M11*Elt.M23 - Elt.M13*Elt.M21)*(Elt.M32*Elt.M44 - Elt.M34*Elt.M42)
      +(Elt.M11*Elt.M24 - Elt.M14*Elt.M21)*(Elt.M32*Elt.M43 - Elt.M33*Elt.M42)
      +(Elt.M12*Elt.M23 - Elt.M13*Elt.M22)*(Elt.M31*Elt.M44 - Elt.M34*Elt.M41)
      -(Elt.M12*Elt.M24 - Elt.M14*Elt.M22)*(Elt.M31*Elt.M43 - Elt.M33*Elt.M41)
      +(Elt.M13*Elt.M24 - Elt.M14*Elt.M23)*(Elt.M31*Elt.M42 - Elt.M32*Elt.M41);
  }

  /// Invert matrix (does not check orthogonality)
  void Invert ()
  {
    T det = Determinant();
    if (det == 0.f) return;

    nglMatrix<T> result;

    det = 1.f / det;
    result.Elt.M11 = det * (Elt.M22*(Elt.M33*Elt.M44 - Elt.M34*Elt.M43) + Elt.M23*(Elt.M34*Elt.M42 - Elt.M32*Elt.M44) + Elt.M24*(Elt.M32*Elt.M43 - Elt.M33*Elt.M42));
    result.Elt.M12 = det * (Elt.M32*(Elt.M13*Elt.M44 - Elt.M14*Elt.M43) + Elt.M33*(Elt.M14*Elt.M42 - Elt.M12*Elt.M44) + Elt.M34*(Elt.M12*Elt.M43 - Elt.M13*Elt.M42));
    result.Elt.M13 = det * (Elt.M42*(Elt.M13*Elt.M24 - Elt.M14*Elt.M23) + Elt.M43*(Elt.M14*Elt.M22 - Elt.M12*Elt.M24) + Elt.M44*(Elt.M12*Elt.M23 - Elt.M13*Elt.M22));
    result.Elt.M14 = det * (Elt.M12*(Elt.M24*Elt.M33 - Elt.M23*Elt.M34) + Elt.M13*(Elt.M22*Elt.M34 - Elt.M24*Elt.M32) + Elt.M14*(Elt.M23*Elt.M32 - Elt.M22*Elt.M33));
    result.Elt.M21 = det * (Elt.M23*(Elt.M31*Elt.M44 - Elt.M34*Elt.M41) + Elt.M24*(Elt.M33*Elt.M41 - Elt.M31*Elt.M43) + Elt.M21*(Elt.M34*Elt.M43 - Elt.M33*Elt.M44));
    result.Elt.M22 = det * (Elt.M33*(Elt.M11*Elt.M44 - Elt.M14*Elt.M41) + Elt.M34*(Elt.M13*Elt.M41 - Elt.M11*Elt.M43) + Elt.M31*(Elt.M14*Elt.M43 - Elt.M13*Elt.M44));
    result.Elt.M23 = det * (Elt.M43*(Elt.M11*Elt.M24 - Elt.M14*Elt.M21) + Elt.M44*(Elt.M13*Elt.M21 - Elt.M11*Elt.M23) + Elt.M41*(Elt.M14*Elt.M23 - Elt.M13*Elt.M24));
    result.Elt.M24 = det * (Elt.M13*(Elt.M24*Elt.M31 - Elt.M21*Elt.M34) + Elt.M14*(Elt.M21*Elt.M33 - Elt.M23*Elt.M31) + Elt.M11*(Elt.M23*Elt.M34 - Elt.M24*Elt.M33));
    result.Elt.M31 = det * (Elt.M24*(Elt.M31*Elt.M42 - Elt.M32*Elt.M41) + Elt.M21*(Elt.M32*Elt.M44 - Elt.M34*Elt.M42) + Elt.M22*(Elt.M34*Elt.M41 - Elt.M31*Elt.M44));
    result.Elt.M32 = det * (Elt.M34*(Elt.M11*Elt.M42 - Elt.M12*Elt.M41) + Elt.M31*(Elt.M12*Elt.M44 - Elt.M14*Elt.M42) + Elt.M32*(Elt.M14*Elt.M41 - Elt.M11*Elt.M44));
    result.Elt.M33 = det * (Elt.M44*(Elt.M11*Elt.M22 - Elt.M12*Elt.M21) + Elt.M41*(Elt.M12*Elt.M24 - Elt.M14*Elt.M22) + Elt.M42*(Elt.M14*Elt.M21 - Elt.M11*Elt.M24));
    result.Elt.M34 = det * (Elt.M14*(Elt.M22*Elt.M31 - Elt.M21*Elt.M32) + Elt.M11*(Elt.M24*Elt.M32 - Elt.M22*Elt.M34) + Elt.M12*(Elt.M21*Elt.M34 - Elt.M24*Elt.M31));
    result.Elt.M41 = det * (Elt.M21*(Elt.M33*Elt.M42 - Elt.M32*Elt.M43) + Elt.M22*(Elt.M31*Elt.M43 - Elt.M33*Elt.M41) + Elt.M23*(Elt.M32*Elt.M41 - Elt.M31*Elt.M42));
    result.Elt.M42 = det * (Elt.M31*(Elt.M13*Elt.M42 - Elt.M12*Elt.M43) + Elt.M32*(Elt.M11*Elt.M43 - Elt.M13*Elt.M41) + Elt.M33*(Elt.M12*Elt.M41 - Elt.M11*Elt.M42));
    result.Elt.M43 = det * (Elt.M41*(Elt.M13*Elt.M22 - Elt.M12*Elt.M23) + Elt.M42*(Elt.M11*Elt.M23 - Elt.M13*Elt.M21) + Elt.M43*(Elt.M12*Elt.M21 - Elt.M11*Elt.M22));
    result.Elt.M44 = det * (Elt.M11*(Elt.M22*Elt.M33 - Elt.M23*Elt.M32) + Elt.M12*(Elt.M23*Elt.M31 - Elt.M21*Elt.M33) + Elt.M13*(Elt.M21*Elt.M32 - Elt.M22*Elt.M31));

    *this = result;
  }

  /// Setup a translation matrix
  void SetTranslation (const nglVector<T>& rDelta)
  {
    SetIdentity();
    Elt.M14 = rDelta[0];
    Elt.M24 = rDelta[1];
    Elt.M34 = rDelta[2];
  }

  /// Setup a translation matrix
  void SetTranslation (T X, T Y, T Z)
  {
    SetIdentity();
    Elt.M14 = X;
    Elt.M24 = Y;
    Elt.M34 = Z;
  }

  /// Compose with a translation
  void Translate (const nglVector<T>& rDelta)
  {
    Elt.M14 += rDelta[0];
    Elt.M24 += rDelta[1];
    Elt.M34 += rDelta[2];
  }

  /// Compose with a translation
  void Translate (T X, T Y, T Z)
  {
    Elt.M14 += X;
    Elt.M24 += Y;
    Elt.M34 += Z;
  }

  /// Setup a scaling transformation (homothetia)
  void SetScaling (const nglVector<T>& rVector)
  {
    Nullify();
    Elt.M11 = rVector[0];
    Elt.M22 = rVector[1];
    Elt.M33 = rVector[2];
    Elt.M44 = 1;
  }

  /// Setup a scaling transformation (homothetia)
  void SetScaling (T X, T Y, T Z)
  {
    Nullify();
    Elt.M11 = X;
    Elt.M22 = Y;
    Elt.M33 = Z;
    Elt.M44 = 1;
  }

  /// Apply a scaling transformation
  void Scale (T X, T Y, T Z)
  {
    Elt.M11 *= X;
    Elt.M21 *= X;
    Elt.M31 *= X;
    Elt.M12 *= Y;
    Elt.M22 *= Y;
    Elt.M32 *= Y;
    Elt.M13 *= Z;
    Elt.M23 *= Z;
    Elt.M33 *= Z;
  }

  /// Apply a scaling transformation
  void Scale (const nglVector<T>& rVector)
  {
    SetScaling(rVector[0], rVector[1], rVector[2]);
  }

  /// Setup a rotation matrix of \a Angle degrees around \a rAxis (using right hand convention)
  void SetRotation (T Angle, const nglVector<T>& rAxis)
  {
    nglVector<T> axis = rAxis;
    axis.Normalize();

    T a = (T) DEG2RAD(Angle);
    T s = (T) sin (a);
    T c = (T) cos (a);
    T t = (T) 1.0 - c;
    T x = axis[0];
    T y = axis[1];
    T z = axis[2];
    T xy = x * y;
    T xz = x * z;
    T yz = y * z;

    Elt.M11 = t * x * x + c;
    Elt.M12 = t * xy    - s * z;
    Elt.M13 = t * xz    + s * y;
    Elt.M14 = 0.f;

    Elt.M21 = t * xy    + s * z;
    Elt.M22 = t * y * y + c;
    Elt.M23 = t * yz    - s * x;
    Elt.M24 = 0.f;

    Elt.M31 = t * xz    - s * y;
    Elt.M32 = t * yz    + s * x;
    Elt.M33 = t * z * z + c;
    Elt.M34 = 0.f;

    Elt.M41 = 0.f;
    Elt.M42 = 0.f;
    Elt.M43 = 0.f;
    Elt.M44 = 1.f;
  }

  /// Setup a rotation matrix of \a Angle degrees around the axis which coordinates are \a X, \a Y and \a Z
  void SetRotation (T Angle, T X, T Y, T Z)
  {
    SetRotation(Angle, nglVector<T>(X, Y, Z));
  }

  /// Compose with a rotation of \a Angle degrees around \a rAxis
  void Rotate (T Angle, const nglVector<T>& rAxis)
  {
    nglMatrix<T> trans;
    trans.SetRotation(Angle, rAxis);
    *this *= trans;
  }

  /// Compose with a rotation of \a Angle degrees around the axis which coordinates are \a X, \a Y and \a Z
  void Rotate (T Angle, T X, T Y, T Z)
  {
    nglMatrix<T> trans;
    trans.SetRotation(Angle, X, Y, Z);
    *this *= trans;
  }

  /// Setup a camera matrix, given eye and target position and up direction
  void LookAt (T EyeX, T EyeY, T EyeZ, T TargetX, T TargetY, T TargetZ, T UpX, T UpY, T UpZ)
  {
    LookAt(nglVector<T>(EyeX, EyeY, EyeZ),
           nglVector<T>(TargetX, TargetY, TargetZ),
           nglVector<T>(UpX, UpY, UpZ));
  }

  /// Setup a camera matrix, given eye and target position and up direction
  void LookAt (const nglVector<T>& rEye, const nglVector<T>& rTarget, const nglVector<T>& rUp)
  {
    nglVector<T> forward = rTarget - rEye;
    forward.Normalize();

    /* Side = forward x up */
    nglVector<T> side = forward ^ rUp;
    side.Normalize();

    /* Recompute up as : up = side x forward */
    nglVector<T> up = side ^ forward;

    Elt.M11 = side[0];
    Elt.M12 = side[1];
    Elt.M13 = side[2];
    Elt.M14 = -rEye[0];

    Elt.M21 = up[0];
    Elt.M22 = up[1];
    Elt.M23 = up[2];
    Elt.M24 = -rEye[1];

    Elt.M31 = -forward[0];
    Elt.M32 = -forward[1];
    Elt.M33 = -forward[2];
    Elt.M34 = -rEye[2];

    Elt.M41 = 0.f;
    Elt.M42 = 0.f;
    Elt.M43 = 0.f;
    Elt.M44 = 1.f;
  }

  /*! Interpret the matrix as a rotation matrix with \a pAlpha, \a pBeta and \a pGama Euler angles
      Only return valid results if the matrix determinant is 1
  */
  void GetEulerAngles (T* pAlpha, T* pBeta, T* pGamma) const
  {
    if (Elt.M21 != 0.f)
    {
      if (pAlpha) *pAlpha = (T)asin(Elt.M32);
      if (pBeta)  *pBeta  = (T)atan2(-Elt.M31, Elt.M33);
      if (pGamma) *pGamma = (T)atan2(-Elt.M21, Elt.M22);
    }
    else
    {
      if (pAlpha) *pAlpha = (T)asin(Elt.M32);
      if (pBeta)  *pBeta  = 0.f;
      if (pGamma) *pGamma = (T)atan2(-Elt.M21,Elt.M11);
    }
  }

  /// Setup an orthogonal projection matrix given a clipping cube
  bool SetOrtho (T Left, T Right, T Bottom, T Top, T Near, T Far)
  {
    T deltax, deltay, deltaz;

    deltax = Right - Left;
    deltay = Top - Bottom;
    deltaz = Far - Near;
    if ((deltax == 0.) || (deltay == 0.f) || (deltaz == 0.f))
      return false;

    Nullify();
    Elt.M11 = (T)(2.f / deltax);
    Elt.M14 = (T)(-(Right + Left) / deltax);
    Elt.M22 = (T)(2.f / deltay);
    Elt.M24 = (T)(-(Top + Bottom) / deltay);
    Elt.M33 = (T)(-2.f / deltaz);
    Elt.M34 = (T)(-(Far + Near) / deltaz);
    Elt.M44 = 1.f;

    return true;
  }

  /// Setup an orthogonal projection matrix, with near and far Z planes set to -1 and 1 respectively
  bool SetOrtho (T Left, T Right, T Bottom, T Top)
  {
    return SetOrtho(Left, Right, Bottom, Top, -1.f, 1.f);
  }

  /*! Setup a perspective projection matrix
    \param FovY field of view in degree, in the vertical (Y) direction
    \param Aspect ratio, ie. width / height using viewport dimensions
    \param Near smaller Z value (positive)
    \param Far greater Z value (positive)
  */
  bool SetPerspective (T FovY, T Aspect, T Near, T Far)
  {
    T sine, cotangent, deltaZ;
    T angle = (T)DEG2RAD(FovY);

    deltaZ = Far - Near;
    sine = (T)sin(angle);
    if ((deltaZ == 0.f) || (sine == 0.f) || (Aspect == 0.f))
      return false;

    cotangent = (T)cos(angle) / sine;

    Nullify();
    Elt.M11 = (T)(cotangent / Aspect);
    Elt.M22 = (T)cotangent;
    Elt.M33 = (T)(-(Far + Near) / deltaZ);
    Elt.M43 = -1.f;
    Elt.M34 = (T)(-2.f * Near * Far / deltaZ);

    return true;
  }

  /*! Setup a perspective projection matrix
    \param Left smaller X value (left plane)
    \param Right greater X value (right plane)
    \param Bottom smaller Y value (bottom plane)
    \param Top greater Y value (top plane)
    \param Near smaller Z value (near plane)
    \param Far greater Z value (far plane)
  */
  bool SetFrustum (T Left, T Right, T Bottom, T Top, T Near, T Far)
  {
    T deltaX, deltaY, deltaZ;

    deltaX = Right - Left;
    deltaY = Top - Bottom;
    deltaZ = Far - Near;
    if ((Near <= 0.f) || (Far <= 0.f) ||
        (deltaX == 0.f) || (deltaY == 0.f) || (deltaZ == 0.f)) 
      return false;

    Nullify();
    Elt.M11 = (T)(Near * 2.f / deltaX);
    Elt.M22 = (T)(Near * 2.f / deltaY);
    Elt.M13 = (T)((Right + Left) / deltaX);
    Elt.M23 = (T)((Top + Bottom) / deltaY);
    Elt.M33 = (T)(-(Far + Near) / deltaZ);
    Elt.M43 = (T)(-1.f);
    Elt.M34 = (T)(((T) -2.f) * Near * Far / deltaZ);

    return true;
  }

/* These methods require OpenGL, we only implement them if it is available.
 * They are only implemented in the float and double specialized versions.
 */
#ifdef GL_VERSION_1_1
  /// Load into OpenGL current matrix (see glMatrixMode() and glLoadMatrix{d,f}())
  void LoadMatrix() const
  {
    NGL_ASSERT(!"nglMatrix::LoadMatrix<T> not implemented");
  }

  /// Retrieve the current OpenGL matrix for the given \a Type
  void GetMatrix(GLenum Type = GL_MODELVIEW_MATRIX)
  {
    NGL_ASSERT(!"nglMatrix::GetMatrix<T> not implemented");
  }

  /// Left multiply with current OpenGL matrix, result on the OpenGL side
  void MultMatrix() const
  {
    NGL_ASSERT(!"nglMatrix::MultMatrix<T> not implemented");
  }
#endif // GL_VERSION_1_1

  /*
   * Operators
   */

  /// Return the value of the element from the i-th line and the j-th column
  T operator () (uint i, uint j) const
  {
    NGL_ASSERT(i < 4 && j < 4);
    return Array[i + 4*j];
  }

  nglMatrix<T> operator = (const nglMatrix<T>& rMatrix)
  {
    memcpy(Array, rMatrix.Array, 16*sizeof(T));
    return *this;
  }

  /*!< Right matrix multiplication. These two lines perform the same operation :
\code
mat1 *= mat2;
mat1 = mat1 * mat2;
\endcode
  */
  void operator *= (const nglMatrix<T>& rMatrix)
  {
    nglMatrix<T> result;

    result.Elt.M11 =(Elt.M11 * rMatrix.Elt.M11 + Elt.M12 * rMatrix.Elt.M21 + Elt.M13 * rMatrix.Elt.M31 + Elt.M14 * rMatrix.Elt.M41);
    result.Elt.M12 =(Elt.M11 * rMatrix.Elt.M12 + Elt.M12 * rMatrix.Elt.M22 + Elt.M13 * rMatrix.Elt.M32 + Elt.M14 * rMatrix.Elt.M42);
    result.Elt.M13 =(Elt.M11 * rMatrix.Elt.M13 + Elt.M12 * rMatrix.Elt.M23 + Elt.M13 * rMatrix.Elt.M33 + Elt.M14 * rMatrix.Elt.M43);
    result.Elt.M14 =(Elt.M11 * rMatrix.Elt.M14 + Elt.M12 * rMatrix.Elt.M24 + Elt.M13 * rMatrix.Elt.M34 + Elt.M14 * rMatrix.Elt.M44);

    result.Elt.M21 =(Elt.M21 * rMatrix.Elt.M11 + Elt.M22 * rMatrix.Elt.M21 + Elt.M23 * rMatrix.Elt.M31 + Elt.M24 * rMatrix.Elt.M41);
    result.Elt.M22 =(Elt.M21 * rMatrix.Elt.M12 + Elt.M22 * rMatrix.Elt.M22 + Elt.M23 * rMatrix.Elt.M32 + Elt.M24 * rMatrix.Elt.M42);
    result.Elt.M23 =(Elt.M21 * rMatrix.Elt.M13 + Elt.M22 * rMatrix.Elt.M23 + Elt.M23 * rMatrix.Elt.M33 + Elt.M24 * rMatrix.Elt.M43);
    result.Elt.M24 =(Elt.M21 * rMatrix.Elt.M14 + Elt.M22 * rMatrix.Elt.M24 + Elt.M23 * rMatrix.Elt.M34 + Elt.M24 * rMatrix.Elt.M44);

    result.Elt.M31 =(Elt.M31 * rMatrix.Elt.M11 + Elt.M32 * rMatrix.Elt.M21 + Elt.M33 * rMatrix.Elt.M31 + Elt.M34 * rMatrix.Elt.M41);
    result.Elt.M32 =(Elt.M31 * rMatrix.Elt.M12 + Elt.M32 * rMatrix.Elt.M22 + Elt.M33 * rMatrix.Elt.M32 + Elt.M34 * rMatrix.Elt.M42);
    result.Elt.M33 =(Elt.M31 * rMatrix.Elt.M13 + Elt.M32 * rMatrix.Elt.M23 + Elt.M33 * rMatrix.Elt.M33 + Elt.M34 * rMatrix.Elt.M43);
    result.Elt.M34 =(Elt.M31 * rMatrix.Elt.M14 + Elt.M32 * rMatrix.Elt.M24 + Elt.M33 * rMatrix.Elt.M34 + Elt.M34 * rMatrix.Elt.M44);
                                                                                                
    result.Elt.M41 =(Elt.M41 * rMatrix.Elt.M11 + Elt.M42 * rMatrix.Elt.M21 + Elt.M43 * rMatrix.Elt.M31 + Elt.M44 * rMatrix.Elt.M41);
    result.Elt.M42 =(Elt.M41 * rMatrix.Elt.M12 + Elt.M42 * rMatrix.Elt.M22 + Elt.M43 * rMatrix.Elt.M32 + Elt.M44 * rMatrix.Elt.M42);
    result.Elt.M43 =(Elt.M41 * rMatrix.Elt.M13 + Elt.M42 * rMatrix.Elt.M23 + Elt.M43 * rMatrix.Elt.M33 + Elt.M44 * rMatrix.Elt.M43);
    result.Elt.M44 =(Elt.M41 * rMatrix.Elt.M14 + Elt.M42 * rMatrix.Elt.M24 + Elt.M43 * rMatrix.Elt.M34 + Elt.M44 * rMatrix.Elt.M44);

    *this = result;
  }

  /// Matrix multiplication, result is rA * rB
  friend nglMatrix<T> operator * (const nglMatrix& rA, const nglMatrix& rB)
  {
    nglMatrix<T> result = rA;
    result *= rB;
    return result;
  }

  /// Left matrix multiplication with a vector, ie. result = matrix * vector
  nglVector<T> operator * (const nglVector<T>& rVector)
  {
    nglVector<T> result;
    T v0, v1, v2, v3;
    uint i;

    v0 = rVector[0];
    v1 = rVector[1];
    v2 = rVector[2];
    v3 = rVector[3];

    for (i = 0; i < 4; i++)
      result[i] = v0*Array[i] + v1*Array[4+i] + v2*Array[8+i] + v3*Array[12+i];

    return result;
  }

  /// Scalar multiplication (each element is multiplied by a scalar)
  void operator *= (T Scalar)
  {
    uint i;
    for (i = 0; i < 16; i++)
      Array[i] *= Scalar;
  }

  /*
   * Debug
   */

  /// Display the matrix content in the log
  void Dump(int Level = 0, const nglChar* pTitle = NULL) const
  {
    uint i;

    if (pTitle)
    {
      NGL_LOG(NGL_DOMAIN_MATH, Level, _T("%s :"), pTitle);
    }
    for (i = 0; i < 4; i++)
    {
      NGL_LOG(NGL_DOMAIN_MATH, Level, _T("[ %8.3f %8.3f %8.3f %8.3f ]"), (T)(*this)(i,0), (T)(*this)(i,1), (T)(*this)(i,2), (T)(*this)(i,3));
    }
  }
};


/*
 * Specializations
 */

/* These methods require OpenGL, we only implement them if it is available
 */
#ifdef GL_VERSION_1_1
inline void nglMatrix<float>::LoadMatrix() const
{
  glLoadMatrixf(Array); 
}

inline void nglMatrix<double>::LoadMatrix() const
{
  glLoadMatrixd(Array); 
}

inline void nglMatrix<float>::GetMatrix(GLenum Type)
{
  glGetFloatv(Type, Array);
}

inline void nglMatrix<double>::GetMatrix(GLenum Type)
{
  glGetDoublev(Type, Array);
}

inline void nglMatrix<float>::MultMatrix() const
{
  glMultMatrixf(Array); 
}

inline void nglMatrix<double>::MultMatrix() const
{
  glMultMatrixd(Array); 
}
#endif // GL_VERSION_1_1


typedef nglMatrix<float> nglMatrixf;
typedef nglMatrix<double> nglMatrixd;

#endif // __nglMatrix_h__
