/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglModule.h
\brief Runtime dynamic library management
*/

#ifndef __nglModule_h__
#define __nglModule_h__

#include "ngl.h"
#include "nglError.h"
#include "nglPath.h"


// Error codes
#define NGL_MODULE_ENONE     0
#define NGL_MODULE_EPATH     1
#define NGL_MODULE_ELOAD     2


//! Runtime dynamic library management
/*!
This class can load a relocatable object file (DLL) at runtime and query
its symbol's addresses.
*/
class NGL_API nglModule : public nglError
{
public:
  /** @name Life cycle */
  //@{
  nglModule(const nglPath& rPath);
  /*!<
    \param rPath relocatable object file path

    The module will not be loaded until Load() is called. If \a rPath is
    a relative path, system specific automatic research mecanisms are used.
  */
  virtual ~nglModule();
  /*!<
    If a module was loaded, automacally unload it (or more exactly, unreference
    it).
  */
  //@}

  /** @name Loading & unloading */
  //@{
  bool Load();      ///< Load the module. Returns true if succesfull
  bool Unload();    ///< Unload the module. Returns true if succesfull
  bool IsLoaded();  ///< Return true if the module is loaded and GetSymbol() is usable
  //@}

  /** @name Symbol retrieval */
  //@{
  void* GetSymbol(const char* pName);
  /*!< Query a symbol's address
    \param pName symbol name
    \return the address of the symbol, NULL if not found
  */
  //@}

protected:
  virtual const nglChar* OnError (uint& rError) const;

private:
  nglPath   mPath;
#ifdef _WIN32_
  HINSTANCE mHandle;
#else
  void*     mHandle;
#endif // _WIN32_
};

#endif // __nglModule_h__
