/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglOFile.h
\brief File as output stream
*/

#ifndef __nglOFile_h__
#define __nglOFile_h__

#include "ngl.h"
#include "nglOStream.h"
#include "nglFile.h"


///< nglOFile open modes, see nglFile for more info
enum nglOFileMode
{
  eOFileCreate = eFileWrite, 
  eOFileModify = eFileModify,
  eOFileAppend = eFileAppend
};


//! Output file stream
/*!
This output stream is based on the nglFile API, and can be used by all objects
based on generic output streams.
*/
class NGL_API nglOFile : public nglOStream
{
public:
  /** @name Life cycle */
  //@{
  nglOFile(nglFile* pFile);
  /*!<
    \param pFile output file

    Build a nglOFile from an existing nglFile object. The \a pFile object is
    not managed by nglOFile, ie. it won't be destroyed when the stream is
    deleted.

    The creation will fail it \a pFile has been created in read-only mode.
  */
  nglOFile(const nglPath& rPath, nglOFileMode Mode);
  /*!<
    \param rPath file's path
    \param Mode write mode (they are mapped to the write modes of nglFileMode)

    See nglFile::nglFile() and nglFileMode for more info
  */
  virtual ~nglOFile();
  //@}

  /** @name State/error methods */
  //@{
  nglStreamState GetState() const;
  uint           GetError() const;
  const nglChar* GetErrorStr(uint Error) const;
  //@}

  /** @name Stream navigation */
  //@{
  nglSize GetPos() const;
  nglSize SetPos (nglSize Where, nglStreamWhence Whence = eStreamFromStart);
  //@}

  /** @name Output methods */
  //@{
  virtual nglSize Write (const void* pData, nglSize WordCount, uint WordSize);
  virtual void Flush();               ///< Immediatly put pending writes on the storage medium (ignored for a read-only file)
  virtual void AutoFlush(bool Auto);  ///< If auto-flush is set, all write operations will return only once the data is effectively written on the storage medium
  //@}
private:
  nglFile* mpFile;
  bool     mOwnFile;
};

#endif // __nglOFile_h__
