/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglOStream.h
\brief Output data stream base class
*/

#ifndef __nglOStream_h__
#define __nglOStream_h__

#include "ngl.h"
#include "nglString.h"
#include "nglStream.h"


//! Base class for output streams
/*!
This is a pure virtual base class for writable streams (see nglOFile for instance).
*/
class NGL_API nglOStream : public nglStream
{
public:
  /** @name Output methods */
  //@{
  virtual nglSize WriteInt8   (const uint8*  pData, nglSize Count = 1); ///< Write \a Count 8-bit integer values, returns number of bytes effectively written
  virtual nglSize WriteInt16  (const uint16* pData, nglSize Count = 1); ///< Write \a Count 16-bit integer values, returns number of 16-bit words effectively written
  virtual nglSize WriteInt32  (const uint32* pData, nglSize Count = 1); ///< Write \a Count 32-bit integer values, returns number of 32-bit words effectively written
  virtual nglSize WriteInt64  (const uint64* pData, nglSize Count = 1); ///< Write \a Count 64-bit integer values, returns number of 64-bit words effectively written
  virtual nglSize WriteFloat  (const float*  pData, nglSize Count = 1); ///< Write \a Count float values (IEEE 754 simple precision, 32 bits), returns number of 32-bit words effectively written
  virtual nglSize WriteDouble (const double* pData, nglSize Count = 1); ///< Write \a Count double values (IEEE 754 double precision, 64 bits), returns number of 64-bit words effectively written
  virtual nglSize WriteText   (const nglChar* pData, nglSize Count = -1);
  /*!< Write some text
    \param pData text data
    \param Count maximum number of chars to write, none if -1 (\p pData must be null-terminated)
    \return number of \e bytes written

    This method uses the current format and encoding for its output (see SetTextFormat()
    and SetTextEncoding()). Please note that the number of \e bytes (not chars) is
    returned.
  */
  virtual nglSize WriteText   (const nglString& rData);
  /*!< Write some text
    \param rData text data
    \return number of \e bytes written

    This method uses the current format and encoding for its output (see SetTextFormat()
    and SetTextEncoding()). Please note that the number of \e bytes (not chars) is
    returned.
  */
  virtual nglSize Write (const void* pData, nglSize WordCount, uint WordSize) = 0;
  /*!<
    Write \a WordCount number of \a WordSize bytes long from \a pBuffer,
    returns the number of words effectively written.
  */
  //@}

private:
  bool GetConv();
};

#endif // __nglOStream_h__
