/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglPath.h
\brief Portable file system access
*/


#ifndef __nglPath_h__
#define __nglPath_h__

#include "ngl.h"
#include "nglError.h"
#include "nglEndian.h"
#include "nglTime.h"
#include <list>
#ifdef _UNIX_
#include <map>
#endif // _UNIX_

#ifdef _CARBON_
#include <map>
#endif // _CARBON_


// Error codes
#define NGL_PATH_ENONE     0
#define NGL_PATH_EBADPATH  1
#define NGL_PATH_EACCESS   2
#define NGL_PATH_EEXIST    3
#define NGL_PATH_ELEAF     4
#define NGL_PATH_EPROCMAX  5
#define NGL_PATH_ESYSMAX   6
#define NGL_PATH_ETOOLONG  7
#define NGL_PATH_ENOSPACE  8
#define NGL_PATH_EINVALID  9


//! Path base
/*!
Used by nglPath(nglPathBase) to fetch a system dependent path.
*/
enum nglPathBase
{
  ePathCurrent,   ///< Use the current working directory as root (note: the cwd cannot be changed by the application)
  ePathUser,      ///< Use a user-related ('unix sense') location as root
  ePathTemp,      ///< Use a general purpose scratch pad location as root
  ePathApp        ///< This application's binary file path, including the executable file name
};


//! File path information
/*!
A path can be either a leaf (a file, or what can be emulated as a file such as
pipes, device nodes and so on) or a non-leaf (a folder). All fields are only valid
if \a Exists is true. The \a Size value is always zero when the node is not a
leaf.
*/
class nglPathInfo
{
public:
  bool    Exists;      ///< Node existence. All other fields are invalid if set to false.
  bool    IsLeaf;      ///< Leaf (file or assimilable) or non-leaf (folder)
  bool    CanRead;     ///< True if the file (leaf) can be read or the folder (non-leaf) can be traversed and its content listed
  bool    CanWrite;    ///< True if the file (leaf) can be written to or a new node can be created in this folder (non-leaf)
  nglTime LastAccess;  ///< Time stamp of last acess (read or exec)
  nglTime LastMod;     ///< Time stamp of last modification (write)
  nglSize Size;        ///< If the node is a leaf, size in bytes. If non-leaf, always zero.
};


class nglPathVolume;


//! Portable file system access
/*!
Every OS has its own way to handle file systems and its own path name conventions.
The nglPath class tries to make this job more portable (rather than relying simply on
strings), and provide services to explore the host file system.

Some terminology : a nglPath object designates a \e node in the local file system.
The \e node can be either :
- a \e leaf (files and also pipes, sockets, devices, more generally nodes behaving like files)
- a \e non-leaf (folder)
- invalid (an incomplete path name)
*/
class NGL_API nglPath : public nglError
{
public:
  //! Portable charset
  /*!
    All strings passed to nglPath are tested against a valid charset, for
    portability purposes. Build and concatenation operations can display a
    warning via the application log (see nglBaseApp::GetLog) in debug mode.

\code
const char nglPath::ValidChars[] =
  "/.abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890-_";
\endcode
  */
  static const nglChar PortableCharset[];

  /** @name Life cycle */
  //@{
  nglPath();
  /*!<
    nglPath default constructor
  */
#ifdef USE_WCHAR
  nglPath(const char* pPathName);
  /*!< nglPath constructor
    \param pPathName path name, using current locale's encoding

    This method is only available when USE_WCHAR is defined (Unicode mode).
    For portability reasons, any '\\' (anti-slash) characters are replaced by '/' (slash).
  */
#endif
  nglPath(const nglChar* pPathName);
  /*!< nglPath constructor
    \param pPathName path name

    For portability reasons, any '\\' (anti-slash) characters are replaced by '/' (slash).
  */
  nglPath(const nglString& rPathName);
  /*!< nglPath constructor
    \param rPathName path name

    For portability reasons, any '\\' (anti-slash) characters are replaced by '/' (slash).
  */
  nglPath(nglPathBase Base);
  /*!< nglPath constructor
    \param Base path base (see nglPathBase)

    Build a path from a system-dependent location, see nglPathBase.
  */
  nglPath(const nglPath& rPath);
  /*!<
    nglPath copy constructor
  */
  virtual ~nglPath();
  //@}

  /** @name Node operations */
  //@{
  bool Delete(bool Recurse = false);
  /*!< Delete the current node, and optionally its children if any
    \param Recurse if the node is a folder, delete all its content (files and folders) recursively
    \return True if the operation completed successfully. A false return in recursive mode
      means that the operation might have partially succeeded (zero or more nodes were deleted).

    Can raise errors NGL_PATH_EACCESS, NGL_PATH_EBADPATH or NGL_PATH_ETOOLONG.
  */
  bool Create(bool Recurse = true);
  /*!< Interpret current path as a directory and create it
    \param Recurse create intermediate folders if necessary
    \return True if the operation completed successfully. A false return in recursive mode
      means that the operation might have partially succeeded (zero or more nodes created).

    Can raise errors NGL_PATH_EBADPATH (all platforms) and a lot more on unices
    (NGL_PATH_EACCESS, NGL_PATH_ETOOLONG, NGL_PATH_ENOSPACE).
  */
  //@}

  /** @name Path information */
  //@{
  bool      Canonize();               ///< Interpret \e . and \e .. pseudo paths to give a univoque path name (does not flatten symlinks)
  nglString GetPathName() const;      ///< Returns the path name as a string (same information as GetChar())
  nglString GetNodeName() const;      ///< Returns the last path component (node name)
  nglString GetExtension() const;     ///< Returns the file extension (without the dot), or an empty string if the path is a folder or a file without extension
  nglString GetParentName() const;    ///< Returns the parent path as a string (remove the last node)
  nglPath   GetParent() const;        ///< Returns the parent path (remove the last node)
  int       GetChildren(std::list<nglPath>* pChildren) const;
  /*!< Get node's children
    \param pChildren if non-null, children will be appended to this list
    \return Number of children. Zero usually means the node was a leaf, check errors with GetError()

    Do not expect any children ordering, you must explicitely apply an alpha sort if so
    you wish. Children path names contain their own parent's path.

    This method can raise errors NGL_PATH_EEXIST or NGL_PATH_ELEAF (all platforms) and a lot
    more on unices (NGL_PATH_EACCESS, NGL_PATH_EPROCMAX and NGL_PATH_ESYSMAX).
  */
  const nglChar* GetChars() const;    ///< Retrieve path name as a nglChar array (suitable to varargs methods)
  operator const nglChar*() const;    ///< Retrieve path name as a nglChar array (suitable to varargs methods)
  //@}

  /** @name Node information */
  //@{
  bool GetInfo (nglPathInfo& rInfo) const;  ///< Returns node info, see nglPathInfo
  bool Exists() const;                      ///< Check existence of the current node
  bool CanRead() const;                     ///< Check read/traverse access on current node
  bool CanWrite() const;                    ///< Check write access on current node
  bool IsLeaf() const;                      ///< Test wether the node is a leaf (file and assimilated) or not (folder). Calls GetInfo() internally.
  nglTime   GetLastAccess() const;          ///< Returns last (read) access time stamp. Calls GetInfo() internally.
  nglTime   GetLastMod() const;             ///< Returns last modification (write) time stamp. Calls GetInfo() internally.
  nglSize   GetSize() const;                ///< Returns node size in bytes (folders always return zero). Calls GetInfo() internally.
  nglString GetMimeType();
  /*!< Returns the standard MIME type description from file extension.
       If the path is not a file or the type is unknown, returns an empty string
  */
  //@}

  /** @name Special (OS dependent) paths */
  //@{
  const char* GetOSPathName() const;    ///< Returns full path name as seen by the OS, using the current locale
  bool        IsAbsolute() const;       ///< Returns wether the path is absolute or relative
  nglPath     GetAbsolutePath() const;  ///< Returns the absolute path name (from file system or volume root)
  static int  GetVolumes(std::list<nglPathVolume>& rVolumes, uint Flags);
  /*!< List physical 'volumes' and locate them on the filesystem
    \param rVolumes volume descriptions will be appended to this list
    \param Flags filter : only list volumes that have at least the bits from \a Flags set.
           Use nglPathVolume::All to fetch all possible volumes.
    \return number of volumes found

    \e Volume is a term to specify special entry points in the host filesystem,
    like hard disk partition roots, CD/DVD-ROM mount points or drive letter, network shares
    and so on.
  */
  //@}

  /** @name Operators */
  //@{
#ifdef USE_WCHAR
  const nglPath& operator=(const char* pSource);       ///< Initialize a path from a string using locale's encoding
#endif
  const nglPath& operator=(const nglChar* pSource);    ///< Initialize a path from a string
  const nglPath& operator=(const nglString& rSource);  ///< Initialize a path from a string
  const nglPath& operator=(const nglPath& rSource);    ///< Copy a path

  const nglPath& operator+=(const nglPath& rAppend);
  /*!< Path concatenation
    \param rAppend path to append

    See nglPath::operator+ for more information.
  */
  const nglPath& operator+=(const nglString& rAppend);
  /*!< Path concatenation
    \param rAppend path to append

    Variant of the operator+=(const nglPath&),
    the right operand will be nglPath(rAppend).
  */
  friend NGL_API nglPath operator+ (const nglPath& rPath, const nglPath& rAppend);
  /*!< Path concatenation
    \param rPath current object
    \param rAppend path to append

    The concatenation operator will always consider the left operand to be a folder.
    nglPath will take care of proper concatenation regarding the '/' separator.
    You cannot use this operator to add an extension to a file name, you must
    use the nglString domain (see GetPathName() and nglPath::nglPath).

    Here is what you must expect (see the examples/file test) :

\code
nglPath("/usr/bin")  + nglPath("perl")  // path name is '/usr/bin/perl'
nglPath("/usr/bin/") + nglPath("perl")  //              '/usr/bin/perl'
nglPath("/usr/bin")  + nglPath("/perl") //              '/usr/bin/perl'
nglPath("/usr/bin/") + nglPath("/perl") //              '/usr/bin/perl'
\endcode
  */
  friend NGL_API nglPath operator+ (const nglPath& rPath, const nglString& rAppend);
  /*!< Path concatenation
    \param rPath current object
    \param rAppend path to append

    Variant of the operator+(const nglPath&, const nglString&),
    the right operand will be nglPath(rAppend).
  */
  friend NGL_API bool operator==(const nglPath& rLeft, const nglPath& rRight);
  /*!< Path comparison
    The comparison is case sensitive, depending on the underlying file system and OS
  */
  friend NGL_API bool operator==(const nglPath& rLeft, const nglString& rRight);
  /*!< Path comparison
    The comparison is case sensitive, depending on the underlying file system and OS
  */
  friend NGL_API bool operator==(const nglString& rLeft, const nglPath& rRight);
  /*!< Path comparison
    The comparison is case sensitive, depending on the underlying file system and OS
  */
  //@}

protected:
  virtual const nglChar* OnError (uint& rError) const;

private:
  nglString   mPathName;
  char*       mpOSPathName;
#if (defined _UNIX_) || (defined _CARBON_)
  typedef std::map<nglString, nglString, nglString::LessFunctor> MimeMap;

  static nglTime mMimeTypeStamp;
  static MimeMap mMimeType;
#endif // _UNIX_ || _CARBON_

#ifdef USE_WCHAR
  bool  InternalSetPath (const char* pPath);
#endif
  bool  InternalSetPath (const nglChar* pPath);
//  bool ValidateChars();
  int   GetRootPart() const;
};


//! Volume description
/*!
Volume description, as returned by nglPath::GetVolumes().
*/
class NGL_API nglPathVolume
{
public:
  //! Volume media type
  enum Type
  {
    eTypeUnknown,
    eTypeFloppy,
    eTypeHD,
    eTypeCD,
    eTypeDVD,
    eTypeZip,
    eTypeTape,
    eTypeNetwork
  };

  typedef uint Flags;

  static const Flags All;       ///< All flags set
  static const Flags ReadOnly;  ///< Volume is read-only
  static const Flags Removable; ///< Volume is removable
  static const Flags Offline;   ///< Volume is not connected/inserted
  static const Flags System;    ///< System internal volume

  nglPath            Path;      ///< Path to access this volume
  nglString          Comment;   ///< Comment (volume label, share comment, etc)
  Flags              TheFlags;  ///< Various properties
  Type               TheType;   ///< Volume media type
};

#endif // __nglPath_h__
