/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglStream.h
\brief Common definitions and base class for NGL I/O streams
*/

#ifndef __nglStream_h__
#define __nglStream_h__

#include "ngl.h"
#include "nglEndian.h"
#include "nglString.h"


//! Stream states
/*!
Current stream state
*/
enum nglStreamState
{
  eStreamNone,   ///< The stream is uninitialized or destroyed
  eStreamReady,  ///< Data is immediatly available
  eStreamEnd,    ///< End of stream
  eStreamWait,   ///< No data currently available
  eStreamError   ///< An error occured (no more available data)
};

//! Seek hint
/*!
Seeking positions and directions for nglStream::SetPos().
*/
enum nglStreamWhence
{
  eStreamFromStart,  ///< Move forward, from the beginning of the stream (offset 0)
  eStreamForward,    ///< Move forward from the current read/write position
  eStreamRewind,     ///< Move backwards from the current read/write position
  eStreamFromEnd     ///< Move backwards from the end of the stream
};


//! Base class for I/O streams
/*!
NGL provides simple yet advanced stream objects. Input and output streams
are handled in the distinct nglIStream and nglOStream classes. All stream
share the same services :

- seeking in stream when possible

- all base types read and write, any endian convention (see nglEndian.h)

- all read and write methods have two useful calling conventions

- high level text read and write, any encoding (conversions done on the fly
  and transparently), efficient line by line extraction
*/
class NGL_API nglStream
{
public:
  /** @name Life cycle */
  //@{
  nglStream();
  virtual ~nglStream();
  //@}

  /** @name Error management */
  //@{
  virtual uint           GetError() const = 0;               ///< Retrieve the current error code
  virtual const nglChar* GetErrorStr() const;                ///< Retrieve the current error message
  virtual const nglChar* GetErrorStr(uint Error) const = 0;  ///< Retrieve error message by code
  //@}

  /** @name Stream navigation */
  //@{
  virtual nglStreamState GetState() const = 0;
  /*!< Return current stream state (see nglStreamState)
  */
  virtual nglSize GetPos() const = 0;
  /*!< Return current stream position (read/write offset) in bytes
    \return stream position in bytes
  */
  virtual nglSize SetPos (nglSize Where, nglStreamWhence Whence = eStreamFromStart) = 0;
  /*!< Change current stream position
    \param Where motion length in bytes
    \param Whence origin and direction
    \return effective motion length in bytes (using \a Whence direction)

    In many case, backward - and even forward - motions will bot be available, hence
    this method will return zero.
  */
  //@}

  /** @name Endianness */
  //@{
  nglEndian GetEndian() const;
  /*!< Get stream's endianness convention
    \return current endianness convention
  */
  void SetEndian(nglEndian Endian);
  /*!<
    \param Endian new endianness convention

    Set stream's endianness. All further read and writes will use the new
    endianness (buffered pending writes are not affected).
  */
  //@}

  /** @name Text encoding and line-ending format */
  //@{
  nglTextEncoding GetTextEncoding();
  /*!< Retrieve current text encoding
    \return current text encoding

    The ReadLine() and WriteText() methods assume text input and output with
    a given text encoding. As a default, it is set to \e eEncodingNative.
  */
  bool SetTextEncoding (nglTextEncoding Encoding);
  /*!< Set current text encoding
    \param Encoding text encoding

    The ReadLine() and WriteText() methods assume text input and output with
    a given text encoding. As a default, it is set to \e eEncodingNative.
  */
  nglTextFormat GetTextFormat();
  /*!< Retrieve current text output format
    \return current text format

    The WriteText() method must convert end of line charaters appropriately,
    depending on the platform or a given convention.
    As a default, it is set to \e eTextNative.
  */
  void SetTextFormat (nglTextFormat Format);
  /*!< Set current text output format
    \return current text format

    The WriteText() method must convert end of line charaters appropriately,
    depending on the platform or a given convention.
    As a default, it is set to \e eTextNative.
  */
  //@}

protected:
  nglEndian       mEndian;
  nglTextFormat   mTextFormat;
  nglTextEncoding mTextEncoding;
  nglStringConv*  mpConv;
};

#endif // __nglStream_h__
