/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglString.h
\brief High level string manipulation class with Unicode support
*/

#ifndef __nglString_h__
#define __nglString_h__

#include "ngl.h"
#include "nglStringConv.h"
#include <stdarg.h>
#include <vector>
#include <functional>


/* Do some libc API flattening
 */
#ifdef _WIN32_
#  define strcasecmp  stricmp
#  define strncasecmp strnicmp
#  define strcmpi     stricmp
#  define strncmpi    strnicmp
#  define snprintf    _snprintf
#  if !(defined vsnprintf)
#    define vsnprintf _vsnprintf
#  endif
#else
#  define strcmpi  strcasecmp
#  define strncmpi strncasecmp
#endif // _WIN32_

/* Handy macros
 */
#define PLURAL(x)    (((x) > 1) ? _T("s") : _T(""))
#define YESNO(x)     ((x) ? _T("yes") : _T("no"))
#define TRUEFALSE(x) ((x) ? _T("true") : _T("false"))
#define ONOFF(x)     ((x) ? _T("on") : _T("off"))


//! Text termination
/*!
qsed by stream operators (file, network, etc).
*/
enum nglTextFormat
{
  eTextNone,  // no ending code
  eTextDOS,   // \r\n (CR LF)
  eTextUnix,  // \n   (LF)
  eTextMac,   // \r   (CR)
  eTextZero,  // \O
#if defined(_UNIX_) || defined(_BEOS_)
  eTextNative = eTextUnix
#endif
#ifdef _WIN32_
  eTextNative = eTextDOS
#endif
#ifdef _CARBON_
  eTextNative = eTextMac
#endif
};


//! High level string manipulation class with Unicode support
/*!
The nglString class serves two purposes : bring a feature rich and high level interface
to manipulate text strings, and support internationalisation easily. The first goal is
achieved via various and expected methods such as Extract(), Replace(), Trim(), Compare()
or Tokenize(). Then internationalisation (known as 'i18n') is supported via the internal
string format of nglString which relies on every platform native Unicode support.

\attention
NGL can be compiled without Unicode support, and \b is currently shipped
with Unicode turned off (build option) until its implementation matures. Please read at
least the last three paragraphs of this introduction.

The internal Unicode encoding of nglString is platform dependent. It is based on the <i>wide char</i>
ISO-C interface : for instance a <i>wide char</i> is 16 bits long on Windows (which uses UCS-2),
and 32 bits under Linux (more precisely with the GNU libc 2.1.x or later, which uses UCS-4).
If you want to know wether 16 bits are not enough or 32 bits are too much (they are actually
both reasonable), visit http://www.cl.cam.ac.uk/~mgk25/unicode.html (Unicode FAQ). In NGL, the
<i>wide char</i> type is represented by the portable \a nglChar type.

For legacy reasons, most text input and output operates with a language and country dependent
encoding. Saving a text document with the most trivial application (say vi or notepad) implies
the choice of a caracter encoding. Most of the time, the encoding has been selected automatically
for you when your OS asked for your country and your language. You don't need to know explicitely
which encoding you are using, only that you are using it. Likewise, since most textual documents
provide no mean to tell its encoding, an application automatically uses your default encoding
to open and interpret its character codes.

Some examples :

- a western european MS Windows will use the 'Code Page 1252' (aka WinLatin1)
- most american or west european unices will use ISO-8859-1 (aka Latin1) or the newer
ISO-8859-15 (aka Latin15 or Latin0)
- russian unix users are likely to use KOI8-R
- some i18n-ready setups use UTF8 (eg. BeOS)

The nglString class will let you convert the internal Unicode representation from and to these
various encodings. The conversion from and to your current encoding is always possible, minus
conversion errors (see Import() and Export() for more info).

You will mainly deal with three kinds of text data :

- <b>Local</b> : what is stored on the hard disk, or occasionnaly on a LAN, but written by
applications running the same \e locale (or more exactly using the same encoding). In most
cases, plain textual data files use the locale's encoding. The shorter form of Import() and
Export() will do these conversions for you. Note that you don't have to use these methods
explicitely, you can ask for automatic conversion on a per-stream basis (see nglStream).

- <b>Remote</b> : the most famous example is the web browser. Most HTML documents are
served with an encoding information, via the \e Content-Type meta information (embedded in
the HTML page, or more classically in the HTTP headers). You must be ready to convert from
very numerous encodings, but also outputs them (when it comes to POST data). You can also
consider an e-mail client, or more generally any application manipulating textual document
which has to cross country and/or language boundaries.

- <b>Inline</b> : or string literals embedded into your source code. Use the \b L
prefix to define wide chars or 'wide' strings :
\code
nglChar  my_char   = L'x';
nglChar* my_string = L"This is a string literal\n";
\endcode
Your IDE will use your locale's default encoding to encode these characters before feeding
the source file to your compiler. If you want to produce portable code, you should only
use ASCII characters (0x00 - 0x7F) in string literals, since most encodings are supersets
of ASCII (they agree on the numbers and latin characters codes for instance). If you want
to embed non-ASCII characters (such as accents) in a portable way, you can provide their
Unicode value directly with the \b \\u escape sequence :
\code
nglChar* my_string = L"La c\u00E9dille du fran\u00E7ais\n";
\endcode
\b Warning : gcc 2.95.x does not support the \\u escape sequence (gcc 3.x does).<br><br>
While it's portable, it's definitively not readable. You shouldn't embed string literals
in your application besides debug and log outputs. Instead you should move application
strings (dialogs, menu items, etc) in a separate and easily maintainable/translatable
ressource. While NGL does not provide such a framework, we encourage you to consider
GNU gettext (see http://www.gnu.org/software/gettext/).

For the ones who don't need (or don't want) to support i18n, you can choose to let nglString
rely on the 'good old' \e char* interface (and thus an internal 8 bits encoding). You'll have to
define the NO_WCHAR macro globally in your build settings and recompile NGL.
In this case the internal encoding format of nglString is the current locale's one.
Thus you can feed GetChars() to any of your OS function with a \e char*
interface. Please note that only 8 bits character sets are supported in this case.

If you want your application to support both the <i>char*</i> and <i>wide char</i> compilation
modes, all you have to do is stick to the \a nglChar and nglString types, and use the _T()
macro in place of the \b L prefix for your string literals :
\code
nglChar my_char = _T('x');
nglString my_string(_T("This is a string literal"));
const nglChar* my_text = _T("This is a string literal");
\endcode

If you need to deal with multi-byte character sets (such as ShiftJIS or Big5), please use the
Unicode support and the Import() and Export() methods. NGL won't work properly with an internal
multi-byte encoding, since there is no trivial relation between byte length and character count.
*/
class NGL_API nglString
{
public:
  static const nglChar Zero;     ///< The usual 'terminal' zero, cannot be embedded \e in a string
  static const nglChar Tab;      ///< Tabulation character (ie. _T('\\t'))
  static const nglChar NewLine;  ///< Newline character (ie. _T('\\n'))

  static nglString Null;
  static nglString Empty;
  static nglString WhiteSpace;
  /*!<
    Characters considered as white spaces (see Trim() and Tokenize())
    It is initialized with \e space, \e tab and \e newline as white spaces.
  */

public:
  /** @name Life cycle */
  //@{
  nglString ();                                    ///< Creates a null string (contains no data). See IsNull()
  nglString (nglChar Char);                        ///< Creates a string with a single char. If \p Char is zero, the string is empty
  nglString (const nglString& rSource);            ///< Copy constructor
  nglString (const nglChar* pSource);              ///< Creates a string from a null-terminated char array
  nglString (const nglChar* pSource, int Length);  ///< Creates a string from a null-terminated char array, copying no more than \p Length chars

#ifdef USE_WCHAR
  nglString (const char* pSource, nglTextEncoding Encoding = eEncodingNative);
#else
  nglString (const char* pSource, nglTextEncoding Encoding);
#endif
  /*!<
    Creates a string, importing from the null-terminated \a pSource byte array,
    using the given \a Encoding. See Import() for more info.
  */
#ifdef USE_WCHAR
  nglString (const char* pSource, int Length, nglTextEncoding Encoding = eEncodingNative);
#else
  nglString (const char* pSource, int Length, nglTextEncoding Encoding);
#endif
  /*!<
    Creates a string, importing no more than \p Length chars from the
    \p pSource string, using the given \p Encoding. See Import() for more info.
  */

  virtual ~nglString();                            ///< Destructor. Release associated data
  //@}

  /** @name String size */
  //@{
  int  GetLength() const;  ///< Returns size in chars. If zero, the string is either \e null or \e empty
  bool IsNull() const;     ///< Returns wether the string contains allocated data or not
  bool IsEmpty() const;    ///< Returns wether the string contains characters or not. Null strings are considered empty.
  //@}

  /** @name Fetch string content */
  //@{
  nglChar        GetChar (int Index) const;  ///< Returns the char at position \p Index. If \p Index is out of range, returns zero. See operator[]()
  nglChar        GetLastChar() const;        ///< Returns last char of string. Returns zero if the string is \e null or \e empty
  const nglChar* GetChars() const;           ///< Returns the string content as a char array. Returns null if the string is \e null

  char*          Export (const nglTextEncoding Encoding = eEncodingNative) const;
  /*!< Export to a given text encoding
    \param Encoding target encoding
    \return encoded string content, null if the string is \e null or an error occured

    This simplified export interface returns a malloc'ed buffer which \e must be
    free'd by the user if non null. A valid output is always a null-terminated byte
    array.

    System functions expect char* strings to be encoded using the current locale's encoding
    ("" empty string). Example :
\code
bool SetEnv(nglString& rName, nglString& rValue, bool Overwrite)
{
  char* name = rName.Export();
  char* value = rValue.Export();
  int ret = -1;

  if (name && value)
    ret = setenv(name, value, Overwrite ? 1 : 0);

  if (value) free(value);
  if (name) free(name);

  return (ret == 0);
}
\endcode
  */
  int            Export (int& rOffset, char* pBuffer, int& rToWrite, const nglTextEncoding Encoding = eEncodingNative) const;
  /*!< Export to a given text encoding
    \param rOffset starting offset in string
    \param pBuffer target buffer
    \param rToWrite number of bytes which can be written at \p pBuffer address
    \param Encoding target encoding
    \return number of conversion errors encountered, or -1 if a fatal error occured

    This method let you provide your own target buffer. This method will write no more than \p rAvailable
    bytes into the \p pBuffer byte array, and will also update \p rOffset and \p rToWrite accordingly to
    reflect the number of bytes effectively read and written.
  */
  int            Export (int& rOffset, char* pBuffer, int& rToWrite, nglStringConv& rConv) const;
  /*!< Export to a given text encoding
    \param rOffset starting offset in string
    \param pBuffer target buffer
    \param rToWrite number of bytes which can be written at \p pBuffer address
    \param rConv conversion context
    \return number of conversion errors encountered, or -1 if a fatal error occured

    Advanced export : this method let you provide your own conversion context (\p rConv)
    and re-use it across multiple calls of Export(). This is usefull to optimize
    chunk oriented conversions of text streams. It is used in the implementation of
    nglOStream.
 */

  nglString      Extract (int Index) const;
  /*!< Fetch a substring
    \param Index start of substring
    \return substring starting from \p Index till end of string.
            Returns a null string if \p Index is negative or below last char
  */
  nglString      Extract (int Index, int Length) const;
  /*!< Fetch a substring
    \param Index start of substring
    \param Length length in chars of substring
    \return \p Length chars from position \p Index.
            Returns a null string if \p Index is negative or below last char or
            if \p Length is negative. \p Length is automatically clamped.
  */
  nglString      GetLeft (int Count) const;  ///< Returns \p Count leftmost chars
  nglString      GetRight(int Count) const;  ///< Returns \p Count rightmost chars

  int            GetInt   (int Base = 10) const;  ///< Parse the string as an integer written in base \p Base
  uint           GetUInt  (int Base = 10) const;  ///< Parse the string as an unsigned integer written in base \p Base
  float          GetFloat () const;               ///< Parse the string as a float decimal representation
  double         GetDouble() const;               ///< Parse the string as a double decimal representation
  
  int            GetCInt   (int Base = 10) const; /*!< Parse the string as an integer written in base \p Base,
                                                       using the 'C' locale (american numeric conventions) */
  uint           GetCUInt  (int Base = 10) const; /*!< Parse the string as an unsigned integer written in base \p Base,
                                                       using the 'C' locale (american numeric conventions) */
  float          GetCFloat () const;              /*!< Parse the string as a float decimal representation,
                                                       using the 'C' locale (american numeric conventions) */
  double         GetCDouble() const;              /*!< Parse the string as a float decimal representation,
                                                       using the 'C' locale (american numeric conventions) */
  //@}

  /** @name Add/modify string content */
  //@{
  bool SetChar (nglChar Char, int Index);
  /*!< Set char value \p Char at position \p Index
    \return false if \p Index is out of range

    A zero char can be inserted, although it is a funny hack (use Delete() instead).
  */

  bool Fill (nglChar Pattern, int RepeatCount);            ///< Fills string with \p RepeatCount chars of value \p Pattern
  bool Fill (const nglChar* pPattern, int RepeatCount);    ///< Fills string with \p RepeatCount occurences of the null-terminated \p pPattern string
  bool Fill (const nglString& rPattern, int RepeatCount);  ///< Fills string with \p RepeatCount occurences of \p rString

  int  Import (const char* pBuffer, const nglTextEncoding Encoding = eEncodingNative);
  /*!< Import text from a given encoding
    \param pBuffer null-terminated string
    \param Encoding source string encoding
    \return number of conversion errors encountered, or -1 if a fatal error occured

    System functions returning char* strings use the locale's current encoding.
    Example :
\code
nglString Home;
Home.Import(getenv("HOME"));
\endcode
  */
  int  Import (const char* pBuffer, int ByteCount, const nglTextEncoding Encoding = eEncodingNative);
  /*!< Import text from a given encoding
    \param pBuffer null-terminated string
    \param ByteCount maximum number of bytes to read
    \param Encoding source string encoding
    \return number of conversion errors encountered, or -1 if a fatal error occured

    This method is identical to Import(const char*, const nglTextEncoding), but let you
    specify a maximum number of bytes to read from \p pBuffer. The conversion still
    stops if a zero is encoutered.
  */
  int  Import (int& rOffset, const char* pBuffer, int& rToRead, const nglTextEncoding Encoding = eEncodingNative);
  /*!< Import text from a given encoding
    \param rOffset start importing from this nglChar offset
    \param pBuffer source string
    \param rToRead source string size in bytes
    \param Encoding source string encoding
    \return number of conversion errors encountered, or -1 if a fatal error occured

    This method is identical to the previous Import() one, but let you specify the source
    buffer size (in bytes), the string start offset (\p rOffset, in nglChars) and update
    both \p rOffset and \p rToRead respectively according to the number of nglChars imported
    and the number of bytes effectively read from \p pBuffer.
  */
  int  Import (int& rOffset, const char* pBuffer, int &rToRead, nglStringConv& rConv);
  /*!< Import text from a given encoding
    \param rOffset start importing from this nglChar offset
    \param pBuffer source string
    \param rToRead source string size in bytes
    \param rConv conversion context
    \return number of conversion errors encountered, or -1 if a fatal error occured

    Advanced import : this method let you provide your own conversion context (\p rConv)
    and re-use it across multiple calls of Import(). This is usefull to optimize
    chunk oriented conversions of text streams. It is used in the implementation of
    nglIStream.
  */

  void Manage (nglChar* pSource);
  /*!< Use \p pSource as string buffer (without copying it)
    Warning, \p pSource will be free()'d at destruction time, call Manage(NULL)
    before destroying the nglString object if you want to prevent this.
  */

  bool Copy   (nglChar Char);              ///< Set current string to one char which value is \p Char
  bool Copy   (const nglChar* pSource);
  /*!< Copy \a pSource content into string
    \param pSource null-terminated string
  */
  bool Copy   (const nglChar* pSource, int SourceLength);   ///< Set current string from null-terminated \p pSource string content, copying no more than \p SourceLength chars
  /*!< Copy no more than \p SourceLength chars from \p pSource into string
    \param pSource null-terminated string
    \param SourceLength maximum number of chars to copy
  */
  bool Copy   (const nglString& rSource);  ///< Copy \p rSource content into string

  bool Insert (nglChar Char, int Index);
  /*!< Insert \p Char at position \p Index
    \return false if \p Index is out of range
  */
  bool Insert (const nglChar* pSource, int Index);
  /*!< Insert the null-terminated \p pSource string at position \p Index
    \return false if \p pSource is null or \p Index is out of range
  */
  bool Insert (const nglString& rSource, int Index);
  /*!< Insert \p rSource content at position \p Index
    \return false if \p rSource is a null string or \p Index is out of range
  */

  bool Append (nglChar Char);              ///< Append \p Char to string
  bool Append (const nglChar* pSource);    ///< Append the null-terminated \p pSource string
  bool Append (const nglString& rSource);  ///< Append \p rSource content

  bool Prepend(nglChar Char);              ///< Prepend \p Char to string
  bool Prepend(const nglChar* pSource);    ///< Prepend the null-terminated \p pSource string
  bool Prepend(const nglString& rSource);  ///< Prepend \p rSource content
  //@}

  /** @name Remove string content */
  //@{
  void Nullify();  ///< Set to null string
  bool Wipe();     ///< Empty string (ie. set to "").

  bool Delete     (int Index = 0);          ///< Delete all chars from position \p Index
  bool Delete     (int Index, int Length);  ///< Delete \p Length chars from position \p Index
  bool DeleteLeft (int Count);              ///< Delete \p Count leftmost chars
  bool DeleteRight(int Count);              ///< Delete \p Count rightmost chars

  bool Trim();                              ///< Remove white spaces from both ends (see WhiteSpace)
  bool Trim (nglChar Char);                 ///< Remove trailing chars \p Char from both ends
  bool Trim (const nglChar* pTrimSet);      ///< Remove chars from the null-termianted \p pTrimSet string from both ends
  bool Trim (const nglString& rTrimSet);    ///< Remove chars from the \p rTrimSet string from both ends

  bool TrimLeft();                             ///< Remove leftmost white spaces (see WhiteSpace)
  bool TrimLeft (nglChar Char);                ///< Remove leftmost trailing chars \p Char
  bool TrimLeft (const nglChar* pTrimSet);     ///< Remove leftmost chars from the null-termianted \p pTrimSet string
  bool TrimLeft (const nglString& rTrimSet);   ///< Remove leftmost chars from the \p rTrimSet string

  bool TrimRight();                            ///< Remove rightmost white spaces (see WhiteSpace)
  bool TrimRight (nglChar Char);               ///< Remove rightmost trailing chars \p Char
  bool TrimRight (const nglChar* pTrimSet);    ///< Remove rightmost chars from the null-termianted \p pTrimSet string
  bool TrimRight (const nglString& rTrimSet);  ///< Remove rightmost chars from the \p rTrimSet string
  //@}

  /** @name Edit string content */
  //@{
  bool Replace (int Index, int Length, nglChar Char);
  /*!< Replace string portion of length \p Length starting from position \p Index with char \p Char
    \return false if \p Index or \p Length are out of range
  */
  bool Replace (int Index, int Length, const nglChar* pSource);
  /*!< Replace string portion of length \p Length starting from position \p Index with the null-terminated \p pSource string
    \return false if \p pSource is null or \p Index or \p Length are out of range
  */
  bool Replace (int Index, int Length, const nglString& rSource);
  /*!< Replace string portion of length \p Length starting from position \p Index with \p rSource
    \return false if \p rSource is a null string or \p Index or \p Length are out of range
  */

  bool Replace (const nglChar Old, const nglChar New);
  /*!< Replace occurences of char \p Old with \p New
  */
  bool Replace (const nglChar* pOld, const nglChar* pNew);
  /*!< Replace occurences of the null-terminated \p pOld string with the null-terminated \p pNew string
    \return false if \p pOld or \p pNew are null
  */
  bool Replace (const nglString& rOld, const nglString& rNew);
  /*!< Replace occurences of \p rOld with \p rNew
    \return false if \p rOld or \p rNew are null strings
  */

  bool ToUpper();                       ///< Convert all lower case chars to upper case
  bool ToUpper(int Index, int Length);  ///< Convert \p Length lower case chars from position \p Index to upper case
  bool ToLower();                       ///< Convert all upper case chars to lower case
  bool ToLower(int Index, int Length);  ///< Convert \p Length upper case chars from position \p Index to lower case

  bool Format (const nglChar* pFormat, ...);
  /*!< Format a string from arguments a la printf
    Please refer to your favorite printf's reference. Make sure you are using
    POSIX or ISOC99 format tags if you want to keep your program portable.

    <b>Important</b> : the %%s tag refers to a Unicode string on \e all platforms
    (ie. _T("hello") or mystring.GetChars()).
  */
  bool Formatv (const nglChar* pFormat, va_list Args);
  /*!< Format a string from arguments a la printf
    This is the vararg version of Format()
  */

  bool CFormat (const nglChar* pFormat, ...);
  /*!< Format a string from arguments a la printf, using the 'C' locale (American numeric conventions) */
  bool CFormatv (const nglChar* pFormat, va_list Args);
  /*!< Format a string from arguments a la printf, using the 'C' locale (American numeric conventions) */

  int HexDump (const char* pBuffer, int ByteCount, bool PrintChar = true, int Columns = 16);
  /*!< Output an hexadecimal byte dump
    \param pBuffer source buffer
    \param ByteCount size of buffer in bytes
    \param PrintChar if true, also display bytes as chars when printable
    \param Columns number of columns per line (one column is one displayed byte)
    \return number of lines

    This method format a byte oriented hexadecimal dump of a given buffer. An example of
    the first 50 bytes of a PNG file as formated by this method with default \p PrintChar
    and \p Columns :
\code
00000000  89 50 4e 47 0d 0a 1a 0a  00 00 00 0d 49 48 44 52  |.PNG........IHDR|
00000010  00 00 00 5a 00 00 00 af  08 02 00 00 00 f2 95 5e  |...Z.........^|
00000020  ff 00 00 20 00 49 44 41  54 78 da 64 bc 69 94 5d  |... .IDATxdi.]|
00000030  d5 75                                             |u              |
\endcode
  */
  //@}

  /** @name String comparison */
  //@{
  int Compare (const nglChar* pSource, bool CaseSensitive = true) const;
  /*!< Compares string with \p pSource as right operand
    \param pSource null-terminated string
    \param CaseSensitive if false perform a case insensitive comparison
    \return 0, -1 or 1 respectively if the string is equal, 'less' or 'greater' than \p pSource
  */
  int Compare (const nglString& rSource, bool CaseSensitive = true) const;
  /*!< Compares string with \p rSource as right operand
    \param rSource right operand string
    \param CaseSensitive if false perform a case insensitive comparison
    \return 0, -1 or 1 respectively if the string is equal, 'less' or 'greater' than \p pSource
  */

  int Compare (const nglChar* pSource, int Index, int Length, bool CaseSensitive = true) const;
  /*!< Compares string portion with \p pSource as right operand
    \param pSource null-terminated string
    \param Index start position of left operand in string
    \param Length length of left operand in string
    \param CaseSensitive if false perform a case insensitive comparison
    \return 0, -1 or 1 respectively if the string portion is equal, 'less' or 'greater' than \p pSource
  */
  int Compare (const nglString& rSource, int Index, int Length, bool CaseSensitive = true) const;
  /*!< Compares string portion with \p pSource as right operand
    \param rSource right operand string
    \param Index start position of left operand in string
    \param Length length of left operand in string
    \param CaseSensitive if false perform a case insensitive comparison
    \return 0, -1 or 1 respectively if the string portion is equal, 'less' or 'greater' than \p pSource
  */
  //@}

  /** @name Substring search */
  //@{
  int Find (nglChar Char, int Index = 0, bool CaseSensitive = true) const;
  /*!< Search the first occurence of char \p Char from position \p Index
    \param Char occurence to find
    \param Index start search from this position
    \param CaseSensitive if false perform a case insensitive comparison
    \return the position of the first occurence, -1 if there is none
  */
  int Find (const nglChar* pSource, int Index = 0, bool CaseSensitive = true) const;
  /*!< Search the first occurence of then null-terminated \p pSource string from position \p Index
    \param pSource occurence to find (null-terminated string)
    \param Index start search from this position
    \param CaseSensitive if false perform a case insensitive comparison
    \return the position of the first occurence, -1 if there is none
  */
  int Find (const nglString& rSource, int Index = 0, bool CaseSensitive = true) const;
  /*!< Search the first occurence of \p rSource from position \p Index
    \param rSource occurence to find
    \param Index start search from this position
    \param CaseSensitive if false perform a case insensitive comparison
    \return the position of the first occurence, -1 if there is none
  */

  int FindLast (nglChar Char, int Index = -1, bool CaseSensitive = true) const;
  /*!< Search the last occurence of char \p Char
    \param Char occurence to find
    \param Index search start index. If negative, start search from GetLength() - Index.
    \param CaseSensitive if false perform a case insensitive comparison
    \return the position of the first occurence, -1 if there is none
  */
  int FindLast (const nglChar* pSource, int Index = -1, bool CaseSensitive = true) const;
  /*!< Search the last occurence of then null-terminated \p pSource string
    \param pSource occurence to find (null-terminated string)
    \param Index search start index. If negative, start search from GetLength() - Index.
    \param CaseSensitive if false perform a case insensitive comparison
    \return the position of the first occurence, -1 if there is none
  */
  int FindLast (const nglString& rSource, int Index = -1, bool CaseSensitive = true) const;
  /*!< Search the last occurence of \p rSource
    \param rSource occurence to find
    \param Index search start index. If negative, start search from GetLength() - Index.
    \param CaseSensitive if false perform a case insensitive comparison
    \return the position of the first occurence, -1 if there is none
  */

  int Contains (nglChar Char, bool CaseSensitive = true) const;              ///< Return the number of occurences of \p Char
  int Contains (const nglChar* pSource, bool CaseSensitive = true) const;    ///< Return the number of occurences of the null-terminated \p pSource string
  int Contains (const nglString& rSource, bool CaseSensitive = true) const;  ///< Return the number of occurences of \p rSource
  //@}

  /** @name Simple tokenizer */
  //@{
  int Tokenize (std::vector<nglString>& rTokens) const;
  /*!< Split a string into tokens, using WhiteSpace as separators
    \param rTokens tokens are appended to this list
    \return number of separators found
  */
  int Tokenize (std::vector<nglString>& rTokens, nglChar Separator) const;
  /*!< Split a string into tokens using \p Separator as separator
    \param rTokens tokens are appended to this list
    \param Separator separator char
    \return number of separators found
  */
  int Tokenize (std::vector<nglString>& rTokens, const nglChar* pSeparators) const;
  /*!< Split a string into tokens using chars from the null-terminated \p pSeparators string as separator
    \param rTokens tokens are appended to this list
    \param pSeparators separators set
    \return number of separators found
  */
  int Tokenize (std::vector<nglString>& rTokens, const nglString& rSeparators) const;
  /*!< Split a string into tokens using chars \p rSeparators as separator
    \param rTokens tokens are appended to this list
    \param rSeparators separators set
    \return number of separators found
  */
/*
  int Tokenize (std::vector <nglString>& rTokens, const nglChar** pSeparatorSet) const;
  int Tokenize (std::vector <nglString>& rTokens, const nglString** pSeparatorSet) const;
*/
  //@}

  /** @name STL functors */
  //@{
  class LessFunctor : public std::binary_function <nglString, nglString, bool> 
  {
  public:
    /*! Comparison functor for STL containers
      \param rLeft left operand
      \param rRight right operand
      \return result of rLeft < rRight

      This functor can be used when nglString is involved in a STL container
      or algorithm. Example :

\code
std::map<nglString, int, nglString::LessFunctor> AgeOfPeople;
nglString someone("John Doe");
int age;

AgeOfPeople[someone] = 33;
age = AgeOfPeople[someone];
\endcode
    */
    bool operator()(const nglString& rLeft, const nglString& rRight) const
    {
      return rLeft < rRight;
    }
  };
  //@}

  /** @name Operators */
  //@{
  // Assignment
  const nglString& operator=(const nglChar Char);
#ifdef USE_WCHAR
  const nglString& operator=(const char* pSource);
#endif
  const nglString& operator=(const nglChar* pSource);
  const nglString& operator=(const nglString& rSource);

  // Appending
  const inline nglString& operator+=(const nglChar Char);
  const inline nglString& operator+=(const nglChar* pSource);
  const inline nglString& operator+=(const nglString& rSource);

  // Concatenation
  friend inline nglString operator+(const nglString& rLeft, const nglChar Right);
  friend inline nglString operator+(const nglString& rLeft, const nglChar* pRight);
  friend inline nglString operator+(const nglString& rLeft, const nglString& rRight);
  friend inline nglString operator+(const nglChar Left, const nglString& rRight);
  friend inline nglString operator+(const nglChar* pLeft, const nglString& rRight);

  // Comparison (equal)
  friend inline bool operator==(const nglString& rLeft, const nglChar Right);
  friend inline bool operator==(const nglString& rLeft, const nglChar* pRight);
  friend inline bool operator==(const nglString& rLeft, const nglString& rRight);
  friend inline bool operator==(const nglChar Left, const nglString& rRight);
  friend inline bool operator==(const nglChar* pLeft, const nglString& rRight);

  // Comparison (different)
  friend inline bool operator!=(const nglString& rLeft, const nglChar Right);
  friend inline bool operator!=(const nglString& rLeft, const nglChar* pRight);
  friend inline bool operator!=(const nglString& rLeft, const nglString& rRight);
  friend inline bool operator!=(const nglChar Left, const nglString& rRight);
  friend inline bool operator!=(const nglChar* pLeft, const nglString& rRight);

  // Comparison (less than)
  friend inline bool operator<(const nglString& rLeft, const nglChar Right);
  friend inline bool operator<(const nglString& rLeft, const nglChar* pRight);
  friend inline bool operator<(const nglString& rLeft, const nglString& rRight);
  friend inline bool operator<(const nglChar Left, const nglString& rRight);
  friend inline bool operator<(const nglChar* pLeft, const nglString& rRight);

  // Comparison (less or equal)
  friend inline bool operator<=(const nglString& rLeft, const nglChar Right);
  friend inline bool operator<=(const nglString& rLeft, const nglChar* pRight);
  friend inline bool operator<=(const nglString& rLeft, const nglString& rRight);
  friend inline bool operator<=(const nglChar Left, const nglString& rRight);
  friend inline bool operator<=(const nglChar* pLeft, const nglString& rRight);

  // Comparison (more than)
  friend inline bool operator>(const nglString& rLeft, const nglChar Right);
  friend inline bool operator>(const nglString& rLeft, const nglChar* pRight);
  friend inline bool operator>(const nglString& rLeft, const nglString& rRight);
  friend inline bool operator>(const nglChar Left, const nglString& rRight);
  friend inline bool operator>(const nglChar* pLeft, const nglString& rRight);

  // Comparison (more or equal)
  friend inline bool operator>=(const nglString& rLeft, const nglChar Right);
  friend inline bool operator>=(const nglString& rLeft, const nglChar* pRight);
  friend inline bool operator>=(const nglString& rLeft, const nglString& rRight);
  friend inline bool operator>=(const nglChar Left, const nglString& rRight);
  friend inline bool operator>=(const nglChar* pLeft, const nglString& rRight);

  // Array access
  inline nglChar operator[](int Index) const;

  // Type cast
  inline operator const nglChar*() const;
  //@}

#ifdef NGL_STRING_STATS
  static void DumpStats(int Level);
#endif

private:
  nglChar* mpBuffer;
  int      mBufferSize;
  int      mLength;
#ifdef NGL_STRING_STATS
  static uint mInstanceCnt;
  static uint mClaimedMem;
  static uint mTotalMem;
  static uint mCurrentMem;
  static uint mMaxMem;
  static uint mAllocCnt;
  static uint mCopyCnt;
  static uint mCopySize;
#endif

  void Init  ();
  bool Resize(int Length, bool Zero = false);
  bool Clamp (int& rIndex, int& rLength) const;
  bool Clamp (int& rIndex, int& rLength, int Max) const;
  bool InternalCopy    (const nglChar* pSource, int SourceLength, int Index, int Length);
  bool InternalFill    (const nglChar* pPattern, int PatternLength, int RepeatCount);
  bool InternalInsert  (const nglChar* pSource, int SourceLength, int Index);
  int  InternalCompare (const nglChar* pSource, int SourceLength, int Index, int Length, bool CaseSensitive) const;
  int  InternalFind    (const nglChar* pSource, int SourceLength, int Index, bool CaseSensitive, bool Forward) const;
  int  InternalContains(const nglChar* pSource, int SourceLength, bool CaseSensitive) const;
  bool InternalReplace (int Index, int Length, const nglChar* pSource, int SourceLength);
  bool InternalReplace (const nglChar* pOld, int OldLength, const nglChar* pNew, int NewLength);
  bool InternalTrim    (const nglChar* pTrimSet, int Sides);
#ifndef _WIN32_
  nglChar* InternalFormatMod (const nglChar* pFormat);
#endif
};


/*
 * Inline code
 */

// Appending to
#ifndef _DOXYGEN_
#define _OP_(x) Append(x); return *this;
#endif
const inline nglString& nglString::operator+=(const nglChar Char)       { _OP_(Char)    }
const inline nglString& nglString::operator+=(const nglChar* pSource)   { _OP_(pSource) }
const inline nglString& nglString::operator+=(const nglString& rSource) { _OP_(rSource) }
#undef _OP_

// Concatenation
#ifndef _DOXYGEN_
#define _OP_(x, y) nglString result(x); result.Append(y); return result;
#endif
inline nglString operator+(const nglString& rLeft, const nglChar Right)     { _OP_(rLeft, Right)  }
inline nglString operator+(const nglString& rLeft, const nglChar* pRight)   { _OP_(rLeft, pRight) }
inline nglString operator+(const nglString& rLeft, const nglString& rRight) { _OP_(rLeft, rRight) }
inline nglString operator+(const nglChar Left, const nglString& rRight)     { _OP_(Left, rRight)  }
inline nglString operator+(const nglChar* pLeft, const nglString& rRight)   { _OP_(pLeft, rRight) }
#undef _OP_

// Comparison (equal)
#ifndef _DOXYGEN_
#define _OP_(x, y) return (x.Compare(y) == 0);
#endif
inline bool operator==(const nglString& rLeft, const nglChar Right)     { _OP_(rLeft, Right)  }
inline bool operator==(const nglString& rLeft, const nglChar* pRight)   { _OP_(rLeft, pRight) }
inline bool operator==(const nglString& rLeft, const nglString& rRight) { _OP_(rLeft, rRight) }
inline bool operator==(const nglChar Left, const nglString& rRight)     { _OP_(rRight, Left)  }
inline bool operator==(const nglChar* pLeft, const nglString& rRight)   { _OP_(rRight, pLeft) }
#undef _OP_

// Comparison (different)
#ifndef _DOXYGEN_
#define _OP_(x, y) return (x.Compare(y) != 0);
#endif
inline bool operator!=(const nglString& rLeft, const nglChar Right)     { _OP_(rLeft, Right)  }
inline bool operator!=(const nglString& rLeft, const nglChar* pRight)   { _OP_(rLeft, pRight) }
inline bool operator!=(const nglString& rLeft, const nglString& rRight) { _OP_(rLeft, rRight) }
inline bool operator!=(const nglChar Left, const nglString& rRight)     { _OP_(rRight, Left)  }
inline bool operator!=(const nglChar* pLeft, const nglString& rRight)   { _OP_(rRight, pLeft) }
#undef _OP_

// Comparison (less than)
#ifndef _DOXYGEN_
#define _OP_(x, y) return (x.Compare(y) < 0);
#endif
inline bool operator<(const nglString& rLeft, const nglChar Right)      { _OP_(rLeft, Right)  }
inline bool operator<(const nglString& rLeft, const nglChar* pRight)    { _OP_(rLeft, pRight) }
inline bool operator<(const nglString& rLeft, const nglString& rRight)  { _OP_(rLeft, rRight) }
#undef _OP_

#ifndef _DOXYGEN_
#define _OP_(x, y) return (x.Compare(y) > 0);
#endif
inline bool operator<(const nglChar Left, const nglString& rRight)      { _OP_(rRight, Left)  }
inline bool operator<(const nglChar* pLeft, const nglString& rRight)    { _OP_(rRight, pLeft) }
#undef _OP_

// Comparison (less or equal)
#ifndef _DOXYGEN_
#define _OP_(x, y) return (x.Compare(y) <= 0);
#endif
inline bool operator<=(const nglString& rLeft, const nglChar Right)     { _OP_(rLeft, Right)  }
inline bool operator<=(const nglString& rLeft, const nglChar* pRight)   { _OP_(rLeft, pRight) }
inline bool operator<=(const nglString& rLeft, const nglString& rRight) { _OP_(rLeft, rRight) }
#undef _OP_

#ifndef _DOXYGEN_
#define _OP_(x, y) return (x.Compare(y) >= 0);
#endif
inline bool operator<=(const nglChar Left, const nglString& rRight)     { _OP_(rRight, Left)  }
inline bool operator<=(const nglChar* pLeft, const nglString& rRight)   { _OP_(rRight, pLeft) }
#undef _OP_

// Comparison (more than)
#ifndef _DOXYGEN_
#define _OP_(x, y) return (x.Compare(y) > 0);
#endif
inline bool operator>(const nglString& rLeft, const nglChar Right)      { _OP_(rLeft, Right)  }
inline bool operator>(const nglString& rLeft, const nglChar* pRight)    { _OP_(rLeft, pRight) }
inline bool operator>(const nglString& rLeft, const nglString& rRight)  { _OP_(rLeft, rRight) }
#undef _OP_

#ifndef _DOXYGEN_
#define _OP_(x, y) return (x.Compare(y) < 0);
#endif
inline bool operator>(const nglChar Left, const nglString& rRight)      { _OP_(rRight, Left)  }
inline bool operator>(const nglChar* pLeft, const nglString& rRight)    { _OP_(rRight, pLeft) }
#undef _OP_

// Comparison (more or equal)
#ifndef _DOXYGEN_
#define _OP_(x, y) return (x.Compare(y) >= 0);
#endif
inline bool operator>=(const nglString& rLeft, const nglChar Right)     { _OP_(rLeft, Right)  }
inline bool operator>=(const nglString& rLeft, const nglChar* pRight)   { _OP_(rLeft, pRight) }
inline bool operator>=(const nglString& rLeft, const nglString& rRight) { _OP_(rLeft, rRight) }
#undef _OP_

#ifndef _DOXYGEN_
#define _OP_(x, y) return (x.Compare(y) <= 0);
#endif
inline bool operator>=(const nglChar Left, const nglString& rRight)     { _OP_(rRight, Left)  }
inline bool operator>=(const nglChar* pLeft, const nglString& rRight)   { _OP_(rRight, pLeft) }
#undef _OP_

// Array access
inline nglChar nglString::operator[](int Index) const
{
#ifdef _DEBUG_
  if (IsNull() || (Index < 0) || (Index >= GetLength())) return Zero;
  return mpBuffer[Index];
#else
  // Relaxed checking, still buffer overflow safe
  return (Index >= 0 && Index < mBufferSize) ? mpBuffer[Index] : Zero;
#endif // _DEBUG_
}

// Type cast
inline nglString::operator const nglChar*() const
{
  return mpBuffer;
}

#endif // __nglString_h__
