/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglVector.h
\brief Vectors (2, 3 and 4 dimensions) template classes
*/

#ifndef __nglVector_h__
#define __nglVector_h__

#include "ngl.h"
#include "nglMath.h"
#include "nglApplication.h"


/// Two dimensional vectors
template <class T> class nglVector2
{
public:
  T Elt[2];

  /// Default constructor, set both coordinates to zero
  nglVector2() { Elt[0] = 0.f; Elt[1] = 0.f; }
  /// Initialize with values
  nglVector2(T X, T Y) { Elt[0] = X; Elt[1] = Y; }
  /// Copy constructor
  nglVector2(const nglVector2<T>& rVector) { *this = rVector; }
  /// Destructor
  virtual ~nglVector2() {}

  /// Copy operator
  void operator=(const nglVector2<T>& rVector) { Elt[0] = rVector[0]; Elt[1] = rVector[1]; }

  /// Read-only access to an element
  T operator[](uint i) const { NGL_ASSERT(i < 2); return Elt[i]; }
  /// Read-write access to an element
  T& operator[](uint i) { NGL_ASSERT(i < 2); return Elt[i]; }

  /// Unary minus
  nglVector2<T> operator-() const { nglVector2<T> result(-Elt[0], -Elt[1]); return result; }

  /// Increment operator
  void operator+=(const nglVector2<T>& rVector) { Elt[0] += rVector[0]; Elt[1] += rVector[1]; }
  /// Decrement operator
  void operator-=(const nglVector2<T>& rVector) { Elt[0] -= rVector[0]; Elt[1] -= rVector[1]; }

  /// Multiplication by a scalar (scaling)
  void operator*=(const T & F) { Elt[0] *= F; Elt[1] *= F; }
  /// Scalar product
  T operator*(const nglVector2<T>& rVector) const { return Elt[0]*rVector[0] + Elt[1]*rVector[1]; }

  /// Squared length
  T SquaredLength() const { return Elt[0]*Elt[0] + Elt[1]*Elt[1]; }
  /// Length
  T Length() const { return (T)sqrt(Elt[0]*Elt[0] + Elt[1]*Elt[1]); }
  /// Normalize. Returns false if the length is zero
  bool Normalize() { T l = Length(); if (l != .0f) { Elt[0] /= l; Elt[1] /= l; return true; } return false; }
};


/// Three dimensional vectors
template <class T> class nglVector3
{
public:
  T Elt[3];

  /// Default constructor, set all coordinates to zero
  nglVector3() { Elt[0] = 0.f; Elt[1] = 0.f; Elt[2] = 0.f; }
  /// Initialize with values
  nglVector3(T X, T Y, T Z) { Elt[0] = X; Elt[1] = Y; Elt[2] = Z; }
  /// Copy constructor
  nglVector3(const nglVector3<T>& rVector) { *this = rVector; }
  /// Destructor
  virtual ~nglVector3() {}

  /// Copy operator
  void operator=(const nglVector3<T>& rVector) { Elt[0] = rVector[0]; Elt[1] = rVector[1]; Elt[2] = rVector[2]; }

  /// Read-only access to an element
  T operator[](uint i) const { NGL_ASSERT(i < 3); return Elt[i]; }
  /// Read-write access to an element
  T& operator[](uint i) { NGL_ASSERT(i < 3); return Elt[i]; }

  /// Unary minus
  nglVector3<T> operator-() const { nglVector3<T> result(-Elt[0], -Elt[1], -Elt[2]); return result; }

  /// Increment operator
  void operator+=(const nglVector3<T>& rVector) { Elt[0] += rVector[0]; Elt[1] += rVector[1]; Elt[2] += rVector[2]; }
  /// Decrement operator
  void operator-=(const nglVector3<T>& rVector) { Elt[0] -= rVector[0]; Elt[1] -= rVector[1]; Elt[2] -= rVector[2]; }

  /// Multiplication by a scalar (scaling)
  void operator*=(const T & F) { Elt[0] *= F; Elt[1] *= F; Elt[2] *= F; }
  /// Scalar product
  T operator*(const nglVector3<T>& rVector) const { return Elt[0]*rVector[0] + Elt[1]*rVector[1] + Elt[2]*rVector[2]; }
  /// Vector product
  void operator^=(const nglVector3<T>& rVector) 
  {
    T x, y, z;

    x = Elt[1] * rVector[2] - Elt[2] * rVector[1];
    y = Elt[2] * rVector[0] - Elt[0] * rVector[2];
    z = Elt[0] * rVector[1] - Elt[1] * rVector[0];

    Elt[0] = x;
    Elt[1] = y;
    Elt[2] = z;
  }
  /// Vector product
  nglVector3<T> operator^(const nglVector3<T>& rVector) const
  {
    nglVector3<T> result = *this;
    result ^= rVector;
    return result;
  }

  /// Squared length
  T SquaredLength() const { return Elt[0]*Elt[0] + Elt[1]*Elt[1] + Elt[2]*Elt[2]; }
  /// Length
  T Length() const { return (T)sqrt(Elt[0]*Elt[0] + Elt[1]*Elt[1] + Elt[2]*Elt[2]); }
  /// Normalize. Returns false if the length is zero
  bool Normalize() { T l = Length(); if (l != .0f) { Elt[0] /= l; Elt[1] /= l; Elt[2] /= l; return true; } return false; }
};


/// Four dimensional (homogeneous) vectors. See also nglMatrix.
template <class T> class nglVector
{
public:
  T Elt[4];

  /// Default constructor, set all coordinates to zero
  nglVector()
  { 
    Elt[0] = 0.f;
    Elt[1] = 0.f;
    Elt[2] = 0.f;
    Elt[3] = 0.f;
  }
  /// Initialize with values
  nglVector(T X, T Y, T Z, T W = 1.f)
  {
    Elt[0] = X;
    Elt[1] = Y;
    Elt[2] = Z;
    Elt[3] = W;
  }
  /// Initialize with a two dimensional vector. Third element is zero, fourth is one.
  nglVector<T>(const nglVector2<T>& rVector)
  { 
    *this = rVector; 
  }
  /// Initialize with a two dimensional vector. Fourth element is one.
  nglVector<T>(const nglVector3<T>& rVector)
  { 
    *this = rVector; 
  }
  /// Copy constructor
  nglVector<T>(const nglVector<T>& rVector) ///< Copy contructor.
  { 
    *this = rVector; 
  }
  /// Destructor
  virtual ~nglVector() ///< Destructor.
  {
  }

  /// Copy operator. Third element is set to zero, fourth to one.
  void operator=(const nglVector2<T>& rVector)
  {
    Elt[0] = rVector[0];
    Elt[1] = rVector[1];
    Elt[2] = 0.f; 
    Elt[3] = 1.f;
  }
  /// Copy operator. Fourth element is set one.
  void operator=(const nglVector3<T>& rVector)
  {
    Elt[0] = rVector[0];
    Elt[1] = rVector[1];
    Elt[2] = rVector[2]; 
    Elt[3] = 1.f;
  }
  /// Copy operator
  void operator=(const nglVector<T>& rVector)
  {
    Elt[0] = rVector[0];
    Elt[1] = rVector[1];
    Elt[2] = rVector[2]; 
    Elt[3] = rVector[3];
  }

  /// Read-only access to an element
  T operator[](uint i) const
  {
    NGL_ASSERT(i < 4);
    return Elt[i]; 
  }
  /// Read-write access to an element
  T& operator[](uint i)
  {
    NGL_ASSERT(i < 4);
    return Elt[i]; 
  }

  /// Unary minus
  nglVector<T> operator-() const
  {
    nglVector<T> result(-Elt[0], -Elt[1], -Elt[2]);
    return result;
  }

  /// Increment operator
  void operator+=(const nglVector<T>& rVector)
  { 
    Elt[0] += rVector[0]; 
    Elt[1] += rVector[1]; 
    Elt[2] += rVector[2]; 
    Elt[3] += rVector[3]; 
  }
  /// Decrement operator
  void operator-=(const nglVector<T>& rVector)
  { 
    Elt[0] -= rVector[0]; 
    Elt[1] -= rVector[1]; 
    Elt[2] -= rVector[2];  
    Elt[3] -= rVector[3];
  }
  /// Vector addition
  nglVector<T> operator+ (const nglVector<T>& rVector) const
  {
    nglVector<T> result = *this;
    result += rVector;
    return result;
  }
  /// Vector substraction
  nglVector<T> operator- (const nglVector<T>& rVector) const
  {
    nglVector<T> result = *this;
    result -= rVector;
    return result;
  }

  /// Multiplication by a scalar (scaling)
  void operator*= (T F)
  {
    Elt[0] *= F;
    Elt[1] *= F;
    Elt[2] *= F;
  }
  /// Multiplication by a scalar (scaling)
  friend nglVector<T> operator* (const nglVector<T>& rVector, T F)
  {
    nglVector<T> result = rVector;
    result *= F;
    return result;
  }
  /// Scalar product
  T operator*(const nglVector<T>& rVector) const
  {
    T result;
    result  = Elt[0] * rVector[0];
    result += Elt[1] * rVector[1];
    result += Elt[2] * rVector[2];
    result += Elt[3] * rVector[3];
    return result;
  }
  /// Vector product
  void operator^=(const nglVector<T>& rVector)
  {
    T x, y, z, w;

    x = Elt[1] * rVector[2] - Elt[2] * rVector[1];
    y = Elt[2] * rVector[0] - Elt[0] * rVector[2];
    z = Elt[0] * rVector[1] - Elt[1] * rVector[0];
    w = Elt[3] * rVector[3];

    Elt[0] = x;
    Elt[1] = y;
    Elt[2] = z;
    Elt[3] = w;
  }
  /// Vector product
  nglVector<T> operator^(const nglVector<T>& rVector)
  {
    nglVector<T> result = *this;
    result ^= rVector;
    return result;
  }

  /// Squared length
  T SquaredLength() const
  {
    return Elt[0]*Elt[0] + Elt[1]*Elt[1] + Elt[2]*Elt[2]; 
  }
  /// Length
  T Length() const
  {
    return (T)sqrt(Elt[0]*Elt[0] + Elt[1]*Elt[1] + Elt[2]*Elt[2]); 
  }
  /// Normalize. Returns false if the length is zero
  bool Normalize()
  {
    T l = Length();
    if (l != 0.f)
    {
      Elt[0] /= l;
      Elt[1] /= l;
      Elt[2] /= l;
      return true;
    }
    return false;
  }

  /// Display the vector's scalar values in the log
  void Dump(int Level = 0, const nglChar* pTitle = NULL) const
  {
    NGL_LOG(NGL_DOMAIN_MATH, Level, _T("%s%s[ %8.3f %8.3f %8.3f %8.3f ]"),
      pTitle ? pTitle : _T(""),
      pTitle ? _T(" = ") : _T(""),
      (T)Elt[0], (T)Elt[1], (T)Elt[2], (T)Elt[3]);
  }
};


/*
 * Specializations
 */

typedef nglVector2<float>  nglVector2f;
typedef nglVector3<float>  nglVector3f;
typedef nglVector<float>   nglVectorf;
typedef nglVector2<double> nglVector2d;
typedef nglVector3<double> nglVector3d;
typedef nglVector<double>  nglVectord;

#endif // __nglVector_h__
