/*
  NGL - C++ cross-platform framework for OpenGL based applications
  Copyright (C) 2000-2003 NGL Team

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*!
\file  nglWindow.h
\brief OpenGL viewport and rendering management

This class is not available if the _NOGFX_ symbol is defined.
*/

#ifndef __nglWindow_h__
#define __nglWindow_h__

#include "ngl.h"
#include "nglError.h"
#include "nglEvent.h"
#include "nglKeyboard.h"
#include "nglString.h"
#include "nglTime.h"
#include "nglKernel.h"
#include "nglContext.h"
#include "nglVideoMode.h"

#ifndef _NOGFX_


//! Window flags type
/*!
This type is used during nglWindow construction, see nglWindowInfo.
*/
typedef uint nglWindowFlags;

//! Window event mask type
/*!
Control the reception of keyboard and mouse events. See nglWindow::SetEventMask().
*/
typedef uint nglWindowEventMask;


//! Mouse mode
/*!
Select the metrics returned by mouse events (relative vs. absolute)
*/
enum nglMouseMode
{
  eMouseAbsolute,  ///< Use absolute coordinates (ie. screen relative)
  eMouseRelative   ///< Use relative coordinates (ie. mouse motions)
};

//! Mouse flags type
/*!
This type is used by the polling function nglWindow::GetMouse() and the
nglWindow mouse event callbacks nglWindow::OnMouseClick and
nglWindow::OnMouseUnclick.
*/
typedef uint nglMouseFlags;

//! Mouse info
/*!
Holds mouse status : coordinates and button pressed.
*/
class nglMouseInfo
{
public:
  int           X;      ///< Horizontal coordinate
  int           Y;      ///< Vertical coordinate
  nglMouseFlags Flags;  ///< Buttons state
};


//! Window initial position hint
/*!
This enum is used in the nglWindowInfo structure passed to the nglWindow
constructor.
*/
enum nglWindowPos
{
  eWindowPosUser,   ///< User specified position, see nglWindowInfo X and Y items
  eWindowPosCenter, ///< Center the window on the screen
  eWindowPosMouse,  ///< Center the window on the current mouse pointer location
  eWindowPosAuto    ///< Let the window manager choose a position
};


//! Window programmable state
/*!
This enum is used to determine the action of the nglWindow::SetState() method.
*/
enum nglWindowSetState
{
  eWindowHide,     /*!< Withdraw the window, including its icon on the desktop
                        or in the taskbar. */
  eWindowShow,     /*!< Display a hidden or minize window. */

  eWindowMinimize  /*!< Minimize a hidden or displayed window. The window itself
                        is not displayed, but the user can bring it back with
                        an icon or a taskbar button. */
};


//! Window current state
/*!
This information is returned by the nglWindow::GetState() method. Note that
it is impossible to distinguish a minimized window from a withdrawn
window. You can only tell if the window exist and if it is displayed.
*/
enum nglWindowState
{
  eWindowNone,   ///< The object is not a valid window (construction failed), delete it as soon as possible
  eWindowHidden, ///< The window is not visible (might be minimized or withdrawn, there is no distinction)
  eWindowVisible ///< The window is displayed on the screen
};


//! Predefined cursor shapes
/*!
See nglWindow::SetCursor() and nglWindow::GetCursor()
*/
enum nglCursorType
{
  eCursorNone,      ///< No cursor
  eCursorArrow,     ///< Standard arrow
  eCursorCross,     ///< Crosshair
  eCursorHand,      ///< Hand
  eCursorHelp,      ///< Question mark with an arrow
  eCursorWait,      ///< Hourglass or watch (might be with an arrow)
  eCursorCaret,     ///< Text caret
  eCursorDnD,       ///< Drag'n'drop symbol
  eCursorForbid,    ///< DnD forbidden (slashed circle)
  eCursorMove,      ///< Four way arrow (horizontal and vertical)
  eCursorResize,    ///< Resize symbol
  eCursorResizeNS,  ///< Double arrow pointing North & South
  eCursorResizeWE,  ///< Double arrow pointing West & East
  eCursorResizeN,   ///< Resize from North
  eCursorResizeS,   ///< Resize from South
  eCursorResizeW,   ///< Resize from West
  eCursorResizeE,   ///< Resize from East
  eCursorResizeNW,  ///< Resize from North West corner
  eCursorResizeNE,  ///< Resize from North East corner
  eCursorResizeSW,  ///< Resize from South West corner
  eCursorResizeSE,  ///< Resize from South East corner
};


//! Window OS-specific properties for nglWindow constructor
/*!
This class is an add-on to nglWindowInfo and encapsulates platform specific
data which may be used by nglWindow constructors (eg. embedded windows
in hosting applications).
*/
class nglWindowOSInfo
{
public:
#ifdef _WIN32_
  HWND Parent;        ///< Parent window. If NULL, creates a regular toplevel window.
  HWND WindowHandle;  ///< Parent window. If NULL, creates a regular toplevel window.

  nglWindowOSInfo();
#endif // _WIN32_

#ifdef _CARBON_
  WindowRef WindowHandle;  ///< Parent window. If NULL, creates a regular toplevel window.

  nglWindowOSInfo();
#endif // _CARBON_
};


//! Window properties for nglWindow constructor
/*!
This structure is only used by the nglWindow constructor to provide the initial
properties of the window. See also nglContextInfo which is used in conjunction
with nglWindowInfo to create a nglWindow.

The \a Pos field is mandatory, and sets the initial positioning strategy.
See \link ::nglWindowPos nglWindowPos \endlink for the details.

The \a XPos and \a YPos items are only relevant if \a Pos is set to
\a eWindowPosUser. They are position coordinates in the user's desktop space,
(0, 0) being the top left corner. You can set a position outside the desktop
(values greater than the desktop's own resolution or negative), however
this is generally not a good idea : platforms might behave differently, some
will simply preventing you from positioning a window outside the desktop.

The \a Width and \a Height fields are mandatory. They let you set the size
of the OpenGL rendering surface. These parameters do not take window decorations
(such as border or title bar) into account.

The \a Flags field is mandatory. If you want to use all defaults, use the
nglWindow::NoFlag value. Any combination of these flags can be used. See each
flag own documentation for more info.

If the the \a FullScreen flag is set in \a Flags, the window dimensions should
match a valid video mode (see nglVideoMode::Enum()).
*/
class NGL_API nglWindowInfo
{
public:
  nglWindowInfo(uint UserWidth = 640, uint UserHeight = 480, bool IsFullScreen = false);
  /*!< Simple constructor
    This is equivalent to this code snippet :
\code
Pos       = eWindowPosAuto;
XPos      = 0;
YPos      = 0;
Width     = UserWidth;
Height    = UserHeight;
Flags     = IsFullScreen ? nglWindow::FullScreen : nglWindow::NoFlag;
Events    = AllEvents;
Cursor    = eCursorArrow;
MouseMode = eMouseAbsolute;
KeyRepeat = true;
Title     = App->GetName();
\endcode
    Note that XPos and YPos are defined but unused since automatic window position
    is requested.
  */

  nglWindowPos       Pos;        ///< Initial position hint
  int                XPos;       ///< X position 
  int                YPos;       ///< Y position
  uint               Width;      ///< Rendering area width
  uint               Height;     ///< Rendering area Height
  nglWindowFlags     Flags;      ///< Window flags
  nglWindowEventMask Events;     ///< Event mask
  nglCursorType      Cursor;     ///< Cursor shape
  nglMouseMode       MouseMode;  ///< Mouse mode
  bool               KeyRepeat;  ///< Key repeat
  nglString          Title;      ///< Caption title
  nglWindowOSInfo    OSInfo;     ///< OS specific options (opaque to the user)
};


/*
 * Error codes
 */
#define NGL_WINDOW_ENONE 0  /*!< There is currently no error. */


//! OpenGL viewport and rendering management
/*!
A nglWindow gives you an OpenGL context for your GL work and a rendering window
for this context. You \e must instantiate a nglWindow before calling any OpenGL
functions. Since you can create any number of windows in an application, you
must specify the context of your OpenGL code : even if you have only one window,
enclose the OpenGL calls between BeginSession() and EndSession() calls.

The \a EndSession method implicitly finishes the rendering (calling
\c glFinish()) and triggers a buffer swap if the window is multi-buffered.
If you render to multiple windows in parallel, it is legal to make successive
calls to \a BeginSession without calling the corresponding \a EndSession. In
other words, you can interleave \a BeginSession and \a EndSession calls at will.

Creating a window requires two pieces of information : OpenGL context info and
general window properties (geometry and behaviour). The first info is given
by the nglContextInfo object, while the second one is simply contained in the
nglWindowInfo struct.

Choosing a OpenGL context is not necessarily an easy task, and mainly not a very
portable job. Keep in mind that the features you ask and get on your
configuration might not (and won't) be available on every user's graphics card
and OpenGL driver combination. Please have a closer look at the nglContextInfo
documentation for this purpose.

Filling the nglWindowInfo is pretty simple. Note that you can change most of
the properties later with various methods (excepted the Pos initial position
hint, which only makes sense at construction time).

When you just created a new nglWindow instance, you should do in this order :

- Check creation success with GetError(). There are many reasons why
a window won't spawn at run time, even if the context and global infos are
valid. The error is non-portable, meaning you cannot interpret the error code.
You should only check against NGL_WINDOW_ENONE, and report the error (if any)
with GetErrorStr(). If you encounter an error, you should delete
the nglWindow instance. An invalid nglWindow will silently fail on any call
on one of its method, and won't generate any events.

- Display the window with a call to \link nglWindow::SetState SetState\endlink(
\link ::nglWindowSetState::eWindowShow eWindowShow \endlink ). You can also
display a window in its minimized state with /a eWindowMinimize.
As a default, newly created window are in the
\link ::nglWindowState::eWindowHidden eWindowHidden \endlink state.

To effectively destroy the window and release the OpenGL context, simply
delete the object.

For more info on nglWindow programming, and especially events management, please
consult the tutorial (available online at http://ngl.sf.net/tutorial or offline
in the \c doc/tutorial folder of the distribution package). You may also peek
into the examples source code in \c examples/window/.
*/
class NGL_API nglWindow : public nglContext
{
public:

  static const nglWindowFlags NoFlag;    ///< Use defaults (all flags unset)
  static const nglWindowFlags NoResize;  /*!< The window manager will prevent the user from resizing the window,
                                              however the programmer can still use nglWindow::SetSize with success. */
  static const nglWindowFlags NoBorder;  /*!< The window will have no decorations at all; often used for so called
                                              'splash screens'. This type of window is implicitly not user-resizable
                                              however the NoResize flag can stay (independently) unset. */
  static const nglWindowFlags FullScreen;///< This window is full screen (a video mode switch occurs if necessary)

  static const nglWindowEventMask NoEvents;     ///< Disable all events
  static const nglWindowEventMask MouseEvents;  /*!< Enable mouse events : OnMouseClick(), OnMouseUnclick() and
                                                     OnMouseMove() callbacks are controlled by this flag. */
  static const nglWindowEventMask KeyEvents;    /*!< Enable key events : OnKeyUp() and OnKeyDown() callbacks are
                                                     controlled by this flag. */
  static const nglWindowEventMask AllEvents;    ///< Enable all events

  static const nglMouseFlags ButtonNone;        ///< No button pressed, can be returned by GetMouse()
  static const nglMouseFlags ButtonLeft;        ///< Left button pressed, this is the only button on most Mac
  static const nglMouseFlags ButtonRight;       ///< Right button pressed
  static const nglMouseFlags ButtonMiddle;      ///< Midle button pressed
  static const nglMouseFlags ButtonWheelUp;     ///< Mouse wheel went one step 'up'
  static const nglMouseFlags ButtonWheelDown;   ///< Mouse wheel went one step 'down'
  static const nglMouseFlags ButtonDoubleClick; ///< The current button has been hit consecutively twice in a short delay

  static void ParseArgs (nglContextInfo& rContext, nglWindowInfo& rInfo);
  /*!< Command line options parser
    \param rContext is updated with a proper nglContext description
    \param rInfo is updated with a proper nglWindow description

    This method will parse some predefined arguments from the command line, and automatically
    fill in sensible values for a context and a window description. These options will be
    supported by your application :

    - <i>--size [width]x[height]</i> : set window (or fullscreen) resolution

    - <i>--fullscreen</i> : go fullscreen. Use current desktop resolution if <i>--size</i>
      is not present
  */

  /** @name Life cycle */
  //@{
  nglWindow (uint Width, uint Height, bool FullScreen = false);
  /*!<
    Simple constructor for the lazy ones.
    \param Width window width
    \param Height window height
    \param FullScreen if true, tries to go fullscreen using width*height as resolution

    This constructor assumes a lot of default values, try to get used to the
    complete form (see below).
    - uses default nglContextInfo, see nglContextInfo documentation
    - initial position is \a eWindowPosAuto
    - all window flags are unset (in other words, set to NoFlag)
    - the window title is 'NGL'

    This is equivalent to the following code :

\code
nglWindow* window;
nglContextInfo context; // Fetch default context
nglWindowInfo info;

info.Pos    = eWindowPosAuto;
info.Width  = Width;
info.Height = Height;
info.Flags  = NoFlag;
info.Events = AllEvents;
info.Title  = _T("NGL");

window = new nglWindow (context, info, NULL);
if (window->GetError())
{
  NGL_OUT (_T("Error : %s, going hara-kiri\n"), window->GetErrorStr());
  delete window;
}
\endcode

    If you refer to the nglWindowInfo struct info, you will notice than you
    can start in full screen mode with the current screen resolution if you
    set both \a Width and \a Height to zero, and the \a FullScreen property
    of the context is set.

    \b Important : do not put your window init code in your derived's class
    constructor, use the OnCreation() callback for this purpose. The reason
    being that the system objects might not be ready for use from the constructor.

    \b Important : check GetError() against NGL_WINDOW_ENONE after calling
    the constructor. If there is an error, the object is invalid  and should
    be deleted. Any method call to such an invalid window will silently fail.
  */

  nglWindow (const nglContextInfo& rContext, const nglWindowInfo& rInfo, const nglContext* pShared = NULL);
  /*!<
    Full constructor for the mighty ones.
    \param rContext OpenGL context info, see nglContextInfo
    \param rInfo   Window options, see nglWindowInfo
    \param pShared Share the OpenGL context with *pShared's one.

    if \c pShared is non null, the newly created context will be shared with
    \c *pShared's one. Both windows will have distinct contexts working
    in parallel, but will share display lists and texture names.
  */

  virtual ~nglWindow();
  /*!<
    Immediatly destroy the window and release the OpenGL context.

    \b Important : you should put your window destroy code in the
    OnDestruction() callback rather than in the destructor. This callback
    will be called event if a delete is issued on the nglWindow object.
  */
  //@}

  /** @name Geometry */
  //@{
  uint           GetWidth() const;   ///< Return the current (user area) width
  uint           GetHeight() const;  ///< Return the current (user area) height
  void           GetSize    (uint& rWidth, uint& rHeight) const;
  /*!<
    Get the window (user area) size.
    \param rWidth current width
    \param rHeight current height
  */
  bool           SetSize    (uint Width, uint Height);
  /*!<
    Sets the window user area size.
    \param Width new width
    \param Height new height
    \return true if the window was resized

    In window'ed mode, \a Width and \a Height are always clamped with a minimum
    value of 1. If the window has been succesfuly resized, returns true.

    In fullscreen mode, this method is ignored.
  */
  bool           GetResolution (float& rHorizontal, float& rVertical);
  /*!<
    Retrieve screen resolution in DPI (dots per inch).
    \param rHorizontal horizontal resolution in DPI
    \param rVertical vertical resolution in DPI
    \return true if \p rHorizontal and \p rVertical contain valid information
  */
  //@}

  /** @name Miscelaneous */
  //@{
//  void           GetInfo    (nglWindowInfo& rInfo);

  nglWindowState GetState   () const;
  /*!<
    Returns the current window display state.
    See \link ::nglWindowState nglWindowState \endlink.
  */
  void           SetState   (nglWindowSetState State);
  /*!<
    Change the window display state.
    \param State new state

    All state transitions are supported. For instance, since a window is
    initially created in the \a eWindowHidden mode, it is possible to
    show it directly in minimized mode : simply ask the \a eWindowMinimize
    mode.

    If the window is 'fullscreen', minimizing or withdrawing the window will
    restore the default desktop resolution, while showing it will switch back
    to the right video mode and restore the fullscreen mode.
  */

  nglString      GetTitle   () const;
  /*!<
    Returns the window current title.
  */
  void           SetTitle   (const nglString& rTitle);
  /*!<
    Sets the window current title.
    \param rTitle new title

    It is legal to pass an empty string (""), the window will have no title.
  */

  nglCursorType  GetCursor  () const;
  /*!<
    Returns the current mouse cursor type
  */
  bool           SetCursor  (nglCursorType CursorType);
  /*!<
    Sets the window current cursor
    \param CursorType new cursor
    \return true if the cursor was succesfully changed
  */

  nglWindowEventMask GetEventMask() const;
  /*!<
    Returns the current event mask
  */
  void SetEventMask(nglWindowEventMask Events);
  /*!<
    Sets a new event mask
    \param Events event mask (combination of nglWindowEventMask constants)
  */

  bool GetKeyRepeat() const;
  /*!<
    Returns wether auto key repeat is active
  */
  void SetKeyRepeat(bool AllowRepeat = true);
  /*!<
    Sets key repeat mode
    \param AllowRepeat if false, disable auto key repeat for this window
  */

  nglMouseMode GetMouseMode() const;
  /*!<
    Returns mouse mode
  */
  void SetMouseMode(nglMouseMode Mode);
  /*!<
    Sets current mouse mode
    \param Mode new mouse mode
  */

  const nglWindowOSInfo* GetOSInfo() const;
  /*!<
    Returns OS-specific informations. See nglWindowOSInfo.
  */
  //@}


  /** @name Polling */
  //@{
  void GetMouse  (nglMouseInfo &rInfo, bool Local = true) const;
  /*!<
    Get the current mouse position and button state.
    \param rInfo mouse information holder
    \param Local if true, returns coordinate relative to the top left corner of the window client area;
           If false, returns the mouse coordinates relative to the top left corner of the screen

    If the mouse mode is \a eMouseRelative, the coordinates will always be both zero
    (only the OnMouseMove() callback can provide relative coordinates); the rInfo.Flags
    state is updated in any case.

    Any number of button flags can be set in the rInfo.Flags bit mask, except
    \a ButtonDoubleClick which does not make sense in this context.
  */
  bool IsKeyDown (nglKeyCode Key) const;
  /*!<
    Test a physical key state.
    \param Key key code

    See the <a href="keyboard.html">nglKeyboard layouts description</a> for physical key names.

    The key code cannot be converted to a printable character, since it would
    require the key context (wether a modifier like Alt, Ctl or Apple was hold,
    or a more complex composing method such as used on asiatic layouts). Use
    the OnKeyDown() callback if you want to interpret key strokes for text
    composing purposes.

    When a window looses its focus (see OnDesactivation()), all key states are
    reset (the keys are virtually released at once).
  */
  //@}

  static const nglChar* GetKeyName (nglKeyCode Key);
  /*!<
    Retrieve the human readable name of a physical key
    \param Key key code

    See nglKeyboard.h and the <a href="keyboard.html">nglKeyboard layouts description</a>.
  */


  /** @name Rendering */
  //@{
  void BeginSession();
  /*!<
    Set the window OpenGL context as the current context. All further OpenGL
    calls will refer to this window context.
  */
  void EndSession();
  /*!<
    Finish the rendering on the current OpenGL context (implicitly calls
    \a glFinish()) and swap frame buffers if appropriate).
  */
  void Invalidate();
  /*!<
    Append a paint event to the event queue. It means the OnPaint() callback
    will be called as soon as possible when you give the control back to the
    main loop.
    
    Only one paint event can be queued. Hence you can call Invalidate()
    many times before returning from an event handler, it will only trigger
    one paint event.
  */
  //@}

  /** @name User callbacks */
  //@{
  virtual void OnCreation();
  /*!<
    This method is called when the window object creation succeeded.
    You should put your init code here, and \e not in the constructor.
  */
  virtual void OnDestruction();
  /*!<
    This method is called when the window object is destroyed.
    You should put your destroy code here, and \e not in the destructor.
  */
  virtual void OnActivation();
  /*!<
    This method is called when the window gets the focus, meaning it
    can receive keyboard and mouse events.
  */
  virtual void OnDesactivation();
  /*!<
    This method is called when the window looses the focus, meaning it
    won't receive further keyboard and mouse events.
  */
  virtual void OnClose();
  /*!<
    This method is called when the user wants to close a displayed or
    iconified window.
    
    The window will not be automatically destroyed, it is up to you to decide
    its fate. It is legal to call \c delete \c this, but don't forget to
    signal the holders of a pointer to this object that it is no longer
    available.
  */
  virtual void OnPaint();
  /*!<
    At least a part of the window must be redrawn. The window might have been
    explicitely invalidated by the programmer (see Invalidate()).
  */
  virtual void OnState       (nglWindowState State);
  /*!<
    This method is called when the window display state changes.
    \param State new display state

    State can be either eWindowHidden (the window has been withdrawn or
    iconified) or eWindowVisible (the window has been unhidden or
    un-iconified).
  */
  virtual void OnResize      (uint Width, uint Height);
  /*!<
    This method is called when the user resizes the window (only possible if
    the NoResize flag is unset) or the programmer calls SetSize().
    \param Width new width
    \param Height new height
  */
  virtual void OnKeyDown     (nglKeyCode Key, nglChar Char);
  /*!<
    This method is called as soon as a key is pressed.
    \param Key key code
    \param Char character value, zero if none

    The key code refers to a language and layout independent code. This is
    useful to specify keys that are located physically on a known place
    on the keyboard. See <a href="keyboard.html">nglKeyboard layout>keyboard layouts</a>.
  */
  virtual void OnKeyUp       (nglKeyCode Key, nglChar Char);
  /*!<
    This method is called when a key is released.
    \param Key key code
    \param Char character value, zero if none

    See OnKeyDown() documentation.
  */
  virtual void OnMouseClick  (int X, int Y, nglMouseFlags Button);
  /*!<
    This method is called as soon as a mouse button is pressed.
    \param X current horizontal pointer coordinate
    \param Y current vertical pointer coordinate
    \param Button button pressed

    See GetMouse() for X and Y exact interpretation.

    \a ButtonDoubleClick and \a ButtonTripleClick bits might be set in \a aButton,
    together with the hit button flag.
  */
  virtual void OnMouseUnclick(int X, int Y, nglMouseFlags Button);
  /*!<
    This method is called when a mouse button is released.
    \param X current horizontal pointer coordinate
    \param Y current vertical pointer coordinate
    \param Button button released

    See GetMouse() for X and Y exact interpretation.

    \a ButtonDoubleClick and \a ButtonTripleClick bits cannot be set in \a Button.
  */
  virtual void OnMouseMove   (int X, int Y);
  /*!<
    This method is called when a mouse motion is detected.
    \param X current horizontal pointer coordinate or displacement
    \param Y current vertical pointer coordinate or displacement

    The X and Y items are displacement measures when the \a RelativeMouse flag
    is set, or window relative coordinates otherwise (the top left corner of
    the window is at 0, 0). You can get the current mouse buttons state with
    the GetMouse() method.

    This event is only triggered if the window has the focus (ie. has received
    a OnActivation() event).
  */
  //@}

protected:
  virtual const nglChar* OnError (uint& rError) const;

private:
  nglWindowOSInfo    mOSInfo;
  nglWindowFlags     mFlags;
  nglWindowEventMask mEventMask;
  nglCursorType      mCursor;
  nglMouseMode       mMouseMode;
  bool               mKeyRepeat;
  bool               mpKeyState[NGL_KEY_MAX];

  void CallOnCreation();
  void CallOnDestruction();
  void CallOnActivation();
  void CallOnDesactivation();
  void CallOnClose();
  void CallOnPaint();
  void CallOnState       (nglWindowState State);
  void CallOnResize      (uint Width, uint Height);
  void CallOnKeyDown     (nglKeyCode Key, nglChar Char);
  void CallOnKeyUp       (nglKeyCode Key, nglChar Char);
  void CallOnMouseClick  (int X, int Y, nglMouseFlags Button);
  void CallOnMouseUnclick(int X, int Y, nglMouseFlags Button);
  void CallOnMouseMove   (int X, int Y);

#ifdef _X11_
public:
  Window         GetHandle();
  void           OnXEvent (const XEvent& rXEvent);

private:
  int            mScreen;
  GC             mGC;
  Colormap       mColormap;
  Atom           mWMProtocol;
  Atom           mWMDelete;
  Atom           mHintsAtom;
  Window         mWindow;
  Cursor         mXCursor;
  Cursor         mEmptyCursor;
  XComposeStatus mCompose;
  bool           mInitCalled;
  bool           mConfigureCalled;
  bool           mIsHidden;
  bool           mNeedGrab;
  bool           mWaitExpose;
  uint           mWidth, mHeight;
  int            mLastMouseX, mLastMouseY;
  bool           mWasMapped;
  int            mLastMappedX, mLastMappedY;
  Time           mLastClick;
  uint           mLastButton;
  int            mClickCnt;
  nglVideoMode   mSavedVideoMode;
  nglVideoMode   mVideoMode;
  bool           mFullScreen;

  void InternalInit (const nglContextInfo& rContext, const nglWindowInfo& rInfo, const nglContext* pShared);
  void DoMouseMove  (const XButtonEvent& rEvent);
  void DoMouseClick (const XButtonEvent& rEvent);
  void DoKeyPress   (const XKeyEvent& rEvent);
  void SetDeco      (uint Deco);
  void Grab         (bool IsGrab);
  Cursor GetXCursor (nglCursorType Cursor);
#endif //_X11_

#ifdef _CARBON_
private:
  bool           mInited;
  nglString      mTitle;
  int            mLastMouseX, mLastMouseY;
  nglContext*    mpContext;
  nglVideoMode*  mpLastVMode;
  nglVideoMode*  mpVMode;
  WindowRef      mWindow;
  bool           mIgnoreNextMouseUp;

  void InternalInit (const nglContextInfo& rContextInfo, const nglWindowInfo& rInfo, const nglContext* pShared);
  OSStatus WindowEventHandler (EventHandlerCallRef eventHandlerCallRef, EventRef eventRef, void* userData);
  void OnLoopInit   ();
  void OnEvent      (int Flags);
  void SetDeco      (uint Deco);
  void Grab         (bool IsGrab);

  friend pascal OSStatus WindowEventHandler (EventHandlerCallRef eventHandlerCallRef, EventRef eventRef, void* userData);
  friend class nglApplication;
#endif //_CARBON_

#ifdef _WIN32_
private:
#ifdef _NGL_WINDOWS_HACKS_ // Define _NGL_WINDOWS_HACKS_ if you want to be able to hack the windows. Highly dangerous!!!
protected:
#endif
  static ATOM mAtom;

  HWND         mHWnd;
  static bool  mClassDefined;
  nglVideoMode mSavedVideoMode;
  bool         mFullScreen;
  int          mOldMouseX;
  int          mOldMouseY;
  nglMouseFlags mMouseButtonStatus;
  nglWindowState mState;

  bool InternalInit(const nglContextInfo& rContext, const nglWindowInfo& rInfo, const nglContext* pShared);
  void SetMouseButtonStatus(nglMouseFlags button);
  void UnsetMouseButtonStatus(nglMouseFlags button);
  void DoKey         (bool IsUp, WPARAM wParam, LPARAM lParam);
  void DoMouseUnclick(nglMouseFlags Flags, HWND hWnd, WPARAM wParam, LPARAM lParam);
  void DoMouseClick  (nglMouseFlags Flags, HWND hWnd, WPARAM wParam, LPARAM lParam);
  static void Exit(); ///< Called to unregister the windows class.


  LRESULT WndProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam );
  friend LRESULT CALLBACK WndProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam );
#endif // _WIN32_

};

#endif // !_NOGFX_
#endif // __nglWindow_h__
